#! /usr/bin/python
################################################################################
#
# File:         bib2target.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bib2target.py,v 1.18 2007/08/16 02:53:02 paepcke Exp $
# Description:  Convert different bibliography formats to html, mif, etc.
# Author:       Andreas Paepcke, Stanford University
# Created:      Tue Dec 19 11:10:53 1995
# Modified:     Wed Aug 15 19:40:17 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Wed Aug 15 19:38:31 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Added Version option (-V, -Version)
# Fri Jul 27 07:22:29 2007 (Andreas Paepcke) paepcke@EMIL
#  Added method getRecsImported() to bibConverter class
# Thu Apr 30 09:38:28 1998 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  In BibliographyImport(), made sure permissions in db were set to 'rw'
#  so that record merging would work.
# Thu Feb 27 12:31:16 1997 (Arturo Crespo) paepcke@Haddock.Stanford.EDU
#  In bibliographyImport, changed 'if len(dbIdentifiers) >= i:' to not have the
#  equal sign.
# Thu Feb 27 12:30:34 1997 (Arturo Crespo) paepcke@Haddock.Stanford.EDU
#  Modified conditional at end to distinguish between stand-alone and imported.
# Fri Sep 20 11:36:22 1996 (Andreas Paepcke) paepcke@Walrus
#  In bibliographyImport: took care of when multiple dbs were involved, and
#  the first was an InterBib db that was write-protected.
# Sun Dec 24 14:11:59 1995 (Andreas Paepcke) paepcke@Stingray
#  Changed to allow import into python programs as well as call from
#  the shell.
################################################################################


# Implements: bib2html, bib2mif, bib2endnote, endnote2bib, endnote2html,
# bib2refer, bib2bibtex, bib2interbib:
# Convert/combine BibTex, Endnote (exported as Refer into file with
# extension 'enr', Strict Refer, and InterBib files among each other

# bib2html, bib2mif, bib2endnote, endnote2bib, endnote2html, bib2refer, refer2bibtex, 
# bib2interbib are all linked to this file. We check how we are called and behave
# accordingly. The file may also be imported. This is done by interbib
# cgi-bin scripts. In that case, function 'main' is not called below.

# Input is a set of bibliography files of mixed format (BibTex, Refer and
# InterBib format, right now). Output is an author-sorted merged
# bibliography in one of several target environments. Currently, html, MIF,
# BibTeX, Refer, and InterBib are supported.  Bibliographies may contain
# links which get translated to links in the target environments. They take
# the form:
#
#                <field>:<targetEnv>:<linkInfo>
#
# For example: 'author:www:http://db.stanford.edu/~paepcke' for the WWW. 
# Or 'title:frame:/u/paepcke/foo.frm mammals' for Framemaker links.
# In BibTex bibs these must be in field 'links'. In Refer they must be in
# field %1. Multiple link specs can be combined as in:
#
#   ['title:www:http://db', 'title:frame:/u/paepcke/foo.frm mammals']
#
# or ['title:www:http://db', 'author:www:http://db.stanford.edu/~paepcke]'
#
# Output is to stdout.
# Options:
#   -h  Show usage
#   -H  Show usage and a bit more help
#   -V  Show InterBib version
#   -v  Run verbose. Error, warning,and progress messages. Default is NOT verbose.
#   -q  Suppress warnings; List errors at the end of normal output
#   -s  Suppress errors and warning. Only normal output.
#   -t  A title. For HTML translations, this title appears at the top of
#       the html page. Default is the first bibliography file name.
#   -e  Target environment: html, mif, endnote refer, bibtex, or interbib.
#       Default is html. This is needed only if you want to call bib2target.py
#       from a script. If you come in through the linked-in files (bib2html, etc.)
#       this option is not relevant.
#
# All options have long forms:
#
#   --Version
#   --verbose
#   --quiet
#   --silent
#   --title
#   --env
#   --help
#   --Help

import sys
import os
import getopt
import string
import bibdb
import bibrefgen
import bibcompile
import bibconvert
import bibutils
import bibexport

interbibVersion = "2.0"

class bibConverter:

  def __init__(self,
	       targetEnv = 'html',
	       title = '',
	       outFd = sys.stdout,
	       errorFd = sys.stderr,
	       verbosity = 'quiet'):
    self.errors = []
    self.warnings = []
    self.bibFiles = []
    self.recsImported = 0
    if verbosity == 'silent':
      self.verbose = 0
      self.reportErrors = 0
    elif verbosity == 'quiet':
      self.verbose = 0
      self.reportErrors = 1
    elif verbosity == 'verbose':
      self.verbose = 1
      self.reportErrors = 1
    self.title = title
    self.db = None
    self.legalTargetEnvs = ['html', 'mif']
    self.targetEnv = targetEnv
    self.outFd = outFd
    self.permissions = 'r' # Permissions for db actions. Strict for when used from WWW
    # Don't know why I thought I needed this following 
    # redirction at some point. But it makes piping not work
    # when running standalone, i.e. from the command line.
    # Redirect stderr to stdout during the duration of processing:
    #sys.stdout = errorFd
    self.usage = "Usage: bib2target [-hHVvqs][-t Page Title] [-e html | mif] <inFileName1> <inFileName2>..."
    self.usageAddendum = "Options: (V)ersion: show InterBib version\n" +\
			 "         (v)erbose: show errors, warnings, results, and process info\n" +\
			 "         (q)iet:    show errors and results (default)\n" +\
			 "         (s)ilent:  show results only\n" +\
			 "         (t)itle:   add title; primarily for HTML\n" +\
			 "         (e)nv:     output as html or mif character sets\n" +\
			 "   The long-form options have syntax like: '--verbose'"


  def inputErrorCheck(self):

    # Find out whether this program was called as bib2target, or as the
    # aliases bib2html or bib2mif, etc:
    
    rest, prog = os.path.split(sys.argv[0])

    if prog == 'bib2html':
      self.usage = "Usage: bib2html [-hHVvqs][-t <Page Title>] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'html'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    elif prog == 'bib2mif':
      self.usage = "Usage: bib2mif [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'mif'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'bib2endnote':
      self.usage = "Usage: bib2endnote [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'endnote'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'endnote2bib':
      self.usage = "Usage: endnote2bib [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'bibtex'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'endnote2html':
      self.usage = "Usage: endnote2html [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'html'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'bib2refer':
      self.usage = "Usage: bib2refer [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'refer'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'refer2bib':
      self.usage = "Usage: refer2bib [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'bibtex'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'bib2bibtex':
      self.usage = "Usage: bib2bib [-hHVvqs] <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'bibtex'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    if prog == 'bib2interbib':
      self.usage = "Usage: bib2interbib [-hHVvqs] <InterBibDBName> <inFileName1> <inFileName2>..." +\
		   "\n          (inFiles may be mixed: BibTeX, Endnote, Strict Refer, or InterBib)"
      self.targetEnv = 'interbib'
      # Can, and need to have our intermediate db writeable:
      self.permissions = 'rw'

    try:
      optlist, args = getopt.getopt(sys.argv[1:], 'hHVvqse:t:', ["help", 
							       "Help",
							       "Version",
							       "verbose",
							       "quiet",
							       "silent",
							       "env=", 
							       "title="])
    except getopt.error, info:
      if self.reportErrors:
        bibutils.report("Illegal option: " + str(info))
        bibutils.report(self.usage)
      return(0)

    for opt, arg in optlist:
        if opt in ("-h", "--help"):
	  bibutils.report(self.usage)
	  return(0)
        if opt in ("-H", "--Help"):
	  bibutils.report(self.usage)
	  bibutils.report(self.usageAddendum)
	  return(0)
        if opt in ("-V", "--Version"):
	  bibutils.report("InterBib version: " + interbibVersion)
	  return(0)
        if opt in ("-v", "--verbose"):
	  self.verbose = True
        if opt in ("-q", "--quiet"):
	  if self.verbose:
	    bibutils.report("Cannot use both 'verbose' and 'quiet'")
	    bibutils.report(self.usage)
	    return(0)
	  self.reportErrors = True
        if opt in ("-s", "--silent"):
	  self.reportErrors = False
	  self.verbose = False
        if opt in ("-t", "--target"):
	  self.target = arg
        if opt in ("-e", "--env"):
	  # Target environment only makes sense when using the
	  # facilities of this file from a script, not when
	  # using it in a command line environment:
	  bibutils.report(self.usage)
	  return(0)

    if len(args) < 1:
      if self.reportErrors:
        bibutils.report(self.usage)
      return(0)
    if (self.targetEnv == 'interbib') and len(args) < 2:
      if self.reportErrors:
        bibutils.report(self.usage)
      return(0)

    return(self.checkBibFiles(args))

  def checkBibFiles(self,bibFileNameList):

    if self.targetEnv == 'interbib':
      # Check that first element of the list is an InterBib file name (no extension):
      if not isInterBibFileName(bibFileNameList[0]):
	if self.reportErrors:
	  bibutils.report("First argument must be an InterBib target file name: No extension")
	return(0)
      self.bibFiles.append(bibFileNameList[0])
      bibFileNameList = bibFileNameList[1:]
	
    for bibFile in bibFileNameList:

      # Make sure all env vars and ~ are resolved:
      bibFile = os.path.expanduser(os.path.expandvars(bibFile))

      if not os.path.isfile(bibFile):
        if self.reportErrors:
	  bibutils.report("File '" + bibFile + "' not found.")
	return(0)

      # Check whether bib file is supported:
      bibsys = checkBibFile(bibFile)
      if not bibsys:
	if self.reportErrors:
	  bibutils.report("Only BibTex, Endnote (exported to Refer with file extension 'enr'), Strict Refer, and InterBib supported: '.bib', 'enr', '.ref' or no extension, respectively")
	return(0)

      self.bibFiles.append(bibFile)
    return(1)

  # dbIdentfiers are use to report progress:

  # Returns number of records imported successfully:
  def bibliographyImport(self,
			 duplicatePolicy = bibdb.autoResolveRecordConflicts,
			 dbIdentifiers = []):

    # If we are converting to an InterBib db, make target db persistent on
    # a file w/ the name given as the first in self.bibFiles.  Note that
    # this case (target InterBib) doesn't happen when we come in through
    # the Web. Only used from bibReferInterbib2interbib:

    if self.targetEnv == 'interbib':
      persistenceFileName = os.path.splitext(self.bibFiles[0])[0]
      self.db = bibdb.BibDB(dbStore=persistenceFileName, permissions=self.permissions)
      # Strip the db name from the bibFiles, leaving only the input files:
      self.bibFiles = self.bibFiles[1:]
    else:
      self.db = bibdb.BibDB(permissions=self.permissions)
      self.db.name = 'UnificationDB'

    # If we are combining more than one db, there is a problem if the first
    # one is an InterBib db: in that case the subsequent ones will try to
    # merge into the InterBib db which won't work since we are only allowed
    # to read into it (they are write-protected, so we have no choice). If
    # the first db is Endnote, Refer or BibTeX, there will be an in-memory structure
    # anyway, so the problem doesn't arise. We try to set sharing to 1 whenever
    # possible to gain speed.

    if len(self.bibFiles) > 1 and isInterBibFileName(self.bibFiles[0]):
      self.db.setDbSharing(0)
    else:
      self.db.setDbSharing(1)

    for i in range(len(self.bibFiles)):
      bibFile = self.bibFiles[i]

      if len(dbIdentifiers) > i:
	dbId = dbIdentifiers[i]
      else:
	dbId = ''

      # The db input char set is LaTex by default. Set output char set
      # to Latin1 which is appropriate for X-Windows:

      if isReferFileName(bibFile) or \
	 isInterBibFileName(bibFile) or \
	 isEndnoteFileName(bibFile):

        self.db.inputCharSet = bibconvert.FrameCharSet

      else:
	self.db.inputCharSet = bibconvert.LatexCharSet

      self.db.outputCharSet = bibconvert.Latin1CharSet

      self.db.setPermissions("rw")

      bibsys = checkBibFile(bibFile)
      if self.verbose:
        bibutils.report('\nBeginning to import bibliography ' + dbId)

      self.recsImported = self.recsImported + \
			  self.db.importBibliography(sourceFile = bibFile,
						     bibsystem = bibsys,
						     duplicatePolicy = duplicatePolicy,
						     verbosity= (self.verbose))

      if self.verbose:
        bibutils.report('done importing bibliography ' + dbId)
      self.errors = self.errors + self.db.errors
      self.warnings = self.warnings + self.db.warnings

    # If target is InterBib db on local file sys, make sure we flush by closing:
    if self.targetEnv == 'interbib':
      self.db.close()
      # Re-open a new one:
      self.db = bibdb.BibDB(dbStore=persistenceFileName, permissions=self.permissions)
    return(self.recsImported)


  # We assume that self.db already has all the bibs imported. If
  # resultStandalone is 1, the resulting text will be a fully
  # legal result file (e.g. in the case of HTML, the correct body
  # header and closing tag will be inserted). Otherwise we simply
  # generate code for the bib entries. This is used if the bibliography
  # is only a piece of a larger document being constructed. Note that
  # if resultStandalone is 0, the title will not be inserted. If
  # annotations is 1, any annotations in the original bibliographies
  # will be generated.

  def bib2target(self, resultStandalone = 1, annote=0):
  
    refList = []

    try:

      entries = self.db.values()

      # Get a reference generator object which takes entries and produces
      # platform-independent references:

      rg = bibrefgen.ReferenceGenerator(verbosity = self.verbose)

      # Get an object that compiles platform-independent references to target:

      if self.targetEnv == 'html':
	c = bibcompile.BibHTMLCompiler()

      elif self.targetEnv == 'mif':
	c = bibcompile.BibMIFCompiler()

      elif self.targetEnv == 'bibtex':
	c = bibexport.BibtexExporter( db = self.db, verbosity=self.verbose)
	c.exportBibliography(sys.stdout, sys.stderr)
	return

      elif self.targetEnv == 'strict refer':
	c = bibexport.ReferExporter( db = self.db, verbosity=self.verbose)
	c.exportBibliography(sys.stdout, sys.stderr)
	return

      elif self.targetEnv == 'endnote':
	c = bibexport.EndnoteExporter( db = self.db, verbosity=self.verbose)
	c.exportBibliography(sys.stdout, sys.stderr)
	return 

      elif self.targetEnv == 'refer':
	c = bibexport.ReferExporter( db = self.db, verbosity=self.verbose)
	c.exportBibliography(sys.stdout, sys.stderr)
	return 

      if self.verbose:
	bibutils.report('\nSorting entries...')
      bibutils.sortByLastName(entries)
      if self.verbose:
	bibutils.report('done.')
      if self.verbose:
	bibutils.report('\nFormatting entries...')
	feedbackGadget = bibutils.feedbackGadget(wholeJob = len(entries),
						 feedbackLeader = '...',
						 feedbackTrailer = ' of formatting done.',
						 reportPercentage = 1)
      for entry in entries:
	refList.append(rg.generateRef(entry, 'GenericStyle', annote=annote))
	if self.verbose:
	  feedbackGadget.tick(1)
      if self.verbose:
	bibutils.report('done formatting entries.')
      self.errors = self.errors + rg.errors
      self.warnings = self.warnings + rg.warnings

      # Restore standard out to its default:
      sys.stdout = self.outFd
      if resultStandalone:
	bibutils.report(c.compileRefList(refList, title = self.title))
      else:
	for ref in refList:
	  bibutils.report(c.compileRef(ref), ioStream=sys.stdout)
      self.errors = self.errors + c.errors
      self.warnings = self.warnings + c.warnings
    
    finally:

      # If we are to report errors only, then do that
      # now. If verbose is turned on then we
      # already listed errors along the way:

      if self.reportErrors and not self.verbose:
	if len(self.errors) > 0:
	  bibutils.report("\n")
	  for error in self.errors:
	    bibutils.report(error)

      # If we were either verbose, or reported errors,
      # give a summay or how many errors and warnings,
      # and how many records were imported:

      if self.verbose or self.reportErrors:
	if self.errors:
	  bibutils.report("\n\nTotal Errors: " + str(len(self.errors)))
        else:
	  bibutils.report("\n\nTotal Errors: none")
	if self.warnings:
	  bibutils.report("Total Warnings: " + str(len(self.warnings)))
        else:
	  bibutils.report("Total Warnings: none")
        bibutils.report("Total records: " + str(self.recsImported))

  # Takes a LIST of STRINGS that are to be filenames
  # with bibliographies to import. Used from the Web to
  # init this list. For console operation the parameter
  # processing takes care of initializing this instance var:

  def setbibFiles(self, bibFileNameList):
    self.bibFiles = bibFileNameList

  def getRecsImported(self):
    return(self.recsImported)
  

def isBibtexFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.bib')

def isEndnoteFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.enr')

def isReferFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '.ref')

def isInterBibFileName(fileName):
  return(os.path.splitext(os.path.basename(fileName))[1] == '')

# Return a string with the bibliography format of the file, as derived
# from the file name extension: If no extension, assume file is an InterBib
# db. If it's a '.bib' extension, assume BibTeX. Else, return 0:

def checkBibFile(bibFile):
  components = os.path.splitext(bibFile)
  if components[1] == '.bib':
    return('bibtex')
  elif components[1] == '.enr':
    return('endnote')
  elif components[1] == '.ref':
    return('refer')
  # an InterBib database?:
  elif components[1] == '':
    return('interbib')
  else:
    return(0)

def main(argc, argv):
  myBibConverter = bibConverter()
  if not myBibConverter.inputErrorCheck():
    return(0)

  myBibConverter.bibliographyImport()
  # If we were just supposed to convert all input files to an InterBib db, we're done:

  if myBibConverter.targetEnv == 'interbib':
    return(1)

  myBibConverter.bib2target()

# If we are called from a shell, run now. If we are being imported,
# do nothing:
if __name__ == '__main__':
  main(len(sys.argv), sys.argv)

  
