#!/usr/bin/python
#
# A class for wrapping the WWW Forms Common Gateway Interface (CGI) 
# Michael McLay, NIST  mclay@eeel.nist.gov  6/14/94
# 
# modified by Steve Majewski <sdm7g@Virginia.EDU> 12/5/94 
#
# now maintained as part of the Python distribution

# Several classes to parse the name/value pairs that are passed to 
# a server's CGI by GET, POST or PUT methods by a WWW FORM. This 
# module is based on Mike McLay's original cgi.py after discussing
# changes with him and others on the comp.lang.python newsgroup, and
# at the NIST Python workshop. 
#
# The rationale for changes was:
#    The original FormContent class was almost, but not quite like
#    a dictionary object. Besides adding some extra access methods,
#    it had a values() method with different arguments and semantics
#    from the standard values() method of a mapping object. Also, 
#    it provided several different access methods that may be necessary
#    or useful, but made it a little more confusing to figure out how
#    to use. Also, we wanted to make the most typical cases the simplest
#    and most convenient access methods. ( Most form fields just return
#    a single value, and in practice, a lot of code was just assuming
#    a single value and ignoring all others. On the other hand, the 
#    protocol allows multiple values to be returned. 
#
#  The new base class (FormContentDict) is just like a dictionary.
#  In fact, if you just want a dictionary, all of the stuff that was
#  in __init__ has been extracted into a cgi.parse() function that will
#  return the "raw" dictionary, but having a class allows you to customize 
#  it further. 
#   Mike McLay's original FormContent class is reimplemented as a 
#  subclass of FormContentDict.
#   There are two additional sub-classes, but I'm not yet too sure 
#  whether they are what I want. 
# 

# ANDREAS: Dec '95:
# Extended to handle Netscape's multipart forms that include delivery of
# files to this script.

# Base class FormContentDict works just like it used to, except now some of
# the dictionary entries may contain entire files that were delivered to
# the script.

# New subclass FileStoringFormContentDict instead stores incoming files in
# temp files. The corresponding dictionary entries each contain a list of
# open file objects. These have an additional attribute: file.sourceFileName
# which is the user-end name of the file that is being delivered.
#
# Invoke method cleanUp() on the form to delete all the files.

# For multipart-form delivery the environment var CONTENT_TYPE is set
# something like the following:
#
# CONTENT_TYPE : multipart/form-data; boundary=---------------------------96252642610044
#
# The boundary is used to piece separate the parts of the multipart form. When
# used (as below), it has an extra '--' prepended. The last one has a '--' appended.
# The example below shows two file delivery form fields and two regular text
# input fields. In each case the first was filled out, the second was left blank
# in the form:
#

# Format of multipart-form delivery (the ' ' are really '\015\012' (^J^M)).
# After each part, one of those is appended

#'-----------------------------96252642610044 '
#'Content-Disposition: form-data; name="bibFile"; filename="trash1.txt" '
#'Content-Type: text/plain '
#' '
#'Line1 of file, '
#'Line2 of file, '
#'... '
#'-----------------------------96252642610044 '
#'Content-Disposition: form-data; name="bibFile"; filename="" '
#' '
#' '
#'-----------------------------96252642610044 '
#'Content-Disposition: form-data; name="bibUrl" '
#' '
#'myurl '
#'-----------------------------96252642610044 '
#'Content-Disposition: form-data; name="bibUrl" '
#' '
#' '
#'-----------------------------96252642610044-- '


import string,re,sys,os,urllib, tempfile
# since os.environ may often be used in cgi code, we name it in this module.
from os import environ

# The FormContent constructor creates a dictionary from the name/value pairs
# passed through the CGI interface.


#
#  form['key'] 
#  form.__getitem__('key') 
#  form.has_key('key')
#  form.keys()
#  form.values()
#  form.items()
#  form.dict

class FormContentDict:
	def __init__( self ):
		self.query_string = environ['QUERY_STRING']
		self.dict = {}
		self.boundaryHeader = '--'
		self.boundaryTrailer = '--'
		self.newFormPart = 1
		self.currentInFileName = ''
		self.currentContentType = ''
		self.formPartBoundary = ''
		self.lastFormPartBoundary = ''
		self.parse()
	def __getitem__(self,key):
		return self.dict[key]
	def keys(self):
		return self.dict.keys()
	def has_key(self, key):
		return self.dict.has_key(key)
	def values(self):
		return self.dict.values()
	def items(self):
		return self.dict.items() 
	def __len__( self ):
		return len(self.dict)
	def parse(self):
		"""Parse the query passed in the environment or on stdin"""
		if environ['REQUEST_METHOD'] == 'GET':
		  return self.parse_qs(environ['QUERY_STRING'])
		elif environ['REQUEST_METHOD'] == 'POST':
		  if string.find(environ['CONTENT_TYPE'], 'multipart/form-data') > -1:
		    return(self.parse_multipart_data(sys.stdin))
		  else:
		    qs = sys.stdin.read(string.atoi(environ['CONTENT_LENGTH']))
		    environ['QUERY_STRING'] = qs
	
	def parse_qs(self, qs):
		"""Parse a query given as a string argument"""
		name_value_pairs = string.splitfields(qs, '&')
		for name_value in name_value_pairs:
			nv = string.splitfields(name_value, '=')
			if len(nv) != 2:
				continue
			name = nv[0]
			value = urllib.unquote(re.sub.gsub('+',' ',nv[1]))
			if len(value):
				if self.dict.has_key (name):
					self.dict[name].append(value)
				else:
					self.dict[name] = [value]
		return self.dict


	def parse_multipart_data(self, fd):

          # Find out what the form part separator is for this form:
	  
	  boundarySpec = string.splitfields(environ['CONTENT_TYPE'], '=')
	  if len(boundarySpec) < 2:
	    raise ValueError, 'Content type field bad:' + boundarySpec
	  self.formPartBoundary = self.boundaryHeader +\
				  boundarySpec[1] +\
				  '\015\012'
	  self.lastFormPartBoundary = self.formPartBoundary[:-2] +\
				      self.boundaryTrailer +\
				      '\015\012'

	  # Get the opening form part boundary and throw it away:
          ln = fd.readline()
	  while 1:
	    # New form part. Get the content disposition:
	    ln = fd.readline()
	    if not ln:
	      return(self.dict) # Should not happen.

	    # ln is the disposition as in:
	    # 'Content-Disposition: form-data; name="bibFile1"; filename=""'
	    info = string.splitfields(ln, ';')

	    # Look for the name of the field and other specs of interest in the
	    # disposition:

	    dictKey = ''
	    self.currentInFileName = ''
	    self.currentContentType = ''

	    for nameValuePair in info[1:]:
	      name = string.strip(string.splitfields(nameValuePair, '=')[0]) # e.g. 'name')
	      value = string.strip(string.splitfields(nameValuePair, '=')[1]) # e.g. 'bibFile1'
	      # Strip the surrounding '"':
	      value = value[1:-1]
	      if name == 'name':
		dictKey = value
	      elif name == 'filename':
		self.currentInFileName = value
	    if not dictKey:
	      raise ValueError, "No field name defined in: '" + ln + "'"

	    # Get the content type (if one is specified):
	    self.currentContentType = fd.readline()
            if string.strip(self.currentContentType):   # More than a CR/LF?
              # Eat the delimiting newline:
              fd.readline()
            if self.processFormPart(fd, dictKey):
	      continue
	    else:
	      # This gives subclass FileStoringFormContentDict a chance to flush:
	      self.finishUp()
	      return(self.dict)

	# Process one form part. Return 1 if more form parts are to be processed
	# else return 0. We make sure that the trailing LF/CR added by protocol
	# is not part of the value. dictKey is the key under which this form part
	# will be inserted in the dictionary:
	
	def processFormPart(self, fd, dictKey):
	  self.newFormPart = 1
	  ln = fd.readline()
	  while 1:
	    # Pull in info from stdin till you find the next form part boundary,
	    # the final boundary or EOF (which shouldn't happen):
	    if not ln:
	      return(0)
	    nextLn = fd.readline()
	    if nextLn == self.lastFormPartBoundary:
	      if (not self.newFormPart) or\
		 (len(ln) > 2):
	        # Skip the trailing LF/CR:
  	        self.processOneLine(dictKey, ln[:-2])
	      return(0)
	    elif nextLn == self.formPartBoundary:
	      # If we just started this form, only keep ln if it
	      # was non-empty:
	      if (not self.newFormPart) or\
		 (len(ln) > 2):
	        # Skip the trailing LF/CR:
  	        self.processOneLine(dictKey, ln[:-2])
	      return(1)
	    else:
	      self.processOneLine(dictKey, ln)
	      ln = nextLn

	def processOneLine(self, name, value):
	  if value:
	    if self.dict.has_key(name):
	      if self.newFormPart:
	        self.dict[name].append(value)
	      else:
		self.dict[name][-1] = self.dict[name][-1] + value
	    else:
	      self.dict[name] = [value]
	  # Acknowledge that we have started to handle a new piece of the form:
  	  self.newFormPart = 0

	def finishUp(self):
	  pass

class FileStoringFormContentDict(FormContentDict):

	def __init__(self):
	  # Initialize the temp file name generation:
	  self.tmpFiles = []
	  # NOTE: this init call below causes all work to be done!:
	  FormContentDict.__init__(self)
    
	def processOneLine(self, name, value):
	  if value:
	    if self.dict.has_key(name):
	      if self.newFormPart:
		if self.currentInFileName:
	          self.dict[name].append(self.startNewFile(value))
		else:
		  self.dict[name].append(value)
	      else:
		if self.currentInFileName:
		  self.dict[name][-1].write(value)
		else:
		  self.dict[name][-1] = self.dict[name][-1] + value
	    else:
	      if self.currentInFileName:
		self.dict[name] = [self.startNewFile(value)]
	      else:
	        self.dict[name] = [value]

	  # Acknowledge that we have started to handle a new piece of the form:
  	  self.newFormPart = 0

        def startNewFile(self, value):
	  file = StrOrFileFd(fileName = tempfile.mktemp(),
			     sourceFileName = self.currentInFileName,
			     mode = 'w')
	  file.write(value)
	  self.tmpFiles.append(file)
	  return(file)
	  
 	def finishUp(self):
	  for file in self.tmpFiles:
	    file.flush()

        def cleanUp(self):
	  for file in self.tmpFiles:
	    os.system('rm ' + file.name())

# This is the "strict" single-value expecting version. 
# IF you only expect a single value for each field, then form[key]
# will return that single value ( the [0]-th ), and raise an 
# IndexError if that expectation is not true. 
# IF you expect a field to have possible multiple values, than you
# can use form.getlist( key ) to get all of the values. 
# values() and items() are a compromise: they return single strings
#  where there is a single value, and lists of strings otherwise. 

class SvFormContentDict(FormContentDict):
	def __getitem__( self, key ):
		if len( self.dict[key] ) > 1 : 
			raise IndexError, 'expecting a single value' 
		return self.dict[key][0]
	def getlist( self, key ):
		return self.dict[key]
	def values( self ):
		lis = []
		for each in self.dict.values() : 
			if len( each ) == 1 : 
				lis.append( each[0] )
			else: lis.append( each )
		return lis
	def items( self ):
		lis = []
		for key,value in self.dict.items():
			if len(value) == 1 :
				lis.append( (key,value[0]) )
			else:	lis.append( (key,value) )
		return lis


# And this sub-class is similar to the above, but it will attempt to 
# interpret numerical values. This is here as mostly as an example,
# but I think the real way to handle typed-data from a form may be
# to make an additional table driver parsing stage that has a table
# of allowed input patterns and the output conversion types - it 
# would signal type-errors on parse, not on access. 
class InterpFormContentDict(SvFormContentDict):
	def __getitem__( self, key ):
		v = SvFormContentDict.__getitem__( self, key )
		if v[0] in string.digits+'+-.' : 
			try:  return  string.atoi( v ) 
			except ValueError:
				try:	return string.atof( v )
				except ValueError: pass
		return string.strip(v)
	def values( self ):
		lis = [] 
		for key in self.keys():
			try:
				lis.append( self[key] )
			except IndexError:
				lis.append( self.dict[key] )
		return lis
	def items( self ):
		lis = [] 
		for key in self.keys():
			try:
				lis.append( (key, self[key]) )
			except IndexError:
				lis.append( (key, self.dict[key]) )
		return lis


# class FormContent parses the name/value pairs that are passed to a
# server's CGI by GET, POST, or PUT methods by a WWW FORM. several 
# specialized FormContent dictionary access methods have been added 
# for convenience.

# function                   return value
#
# form.keys()                     all keys in dictionary
# form.has_key('key')             test keys existance
# form[key]                       returns list associated with key
# form.values('key')              key's list (same as form.[key])
# form.indexed_value('key' index) nth element in key's value list
# form.value(key)                 key's unstripped value 
# form.length(key)                number of elements in key's list
# form.stripped(key)              key's value with whitespace stripped
# form.pars()                     full dictionary 



class FormContent(FormContentDict):
# This is the original FormContent semantics of values,
# not the dictionary like semantics. 
	def values(self,key):
		if self.dict.has_key(key):return self.dict[key]
		else: return None
	def indexed_value(self,key, location):
		if self.dict.has_key(key):
			if len (self.dict[key]) > location:
				return self.dict[key][location]
			else: return None
		else: return None
	def value(self,key):
		if self.dict.has_key(key):return self.dict[key][0]
		else: return None
	def length(self,key):
		return len (self.dict[key])
	def stripped(self,key):
		if self.dict.has_key(key):return string.strip(self.dict[key][0])
		else: return None
	def pars(self):
		return self.dict






def print_environ_usage():
	print """
<H3>These operating system environment variables could have been 
set:</H3> <UL>
<LI>AUTH_TYPE
<LI>CONTENT_LENGTH
<LI>CONTENT_TYPE
<LI>DATE_GMT
<LI>DATE_LOCAL
<LI>DOCUMENT_NAME
<LI>DOCUMENT_ROOT
<LI>DOCUMENT_URI
<LI>GATEWAY_INTERFACE
<LI>LAST_MODIFIED
<LI>PATH
<LI>PATH_INFO
<LI>PATH_TRANSLATED
<LI>QUERY_STRING
<LI>REMOTE_ADDR
<LI>REMOTE_HOST
<LI>REMOTE_IDENT
<LI>REMOTE_USER
<LI>REQUEST_METHOD
<LI>SCRIPT_NAME
<LI>SERVER_NAME
<LI>SERVER_PORT
<LI>SERVER_PROTOCOL
<LI>SERVER_ROOT
<LI>SERVER_SOFTWARE
</UL>
"""

def print_environ():
	skeys = environ.keys()
	skeys.sort()
	print '<h3> The following environment variables ' \
	      'were set by the CGI script: </h3>'
	print '<dl>'
	for key in skeys:
		print '<dt>', escape(key), '<dd>', escape(environ[key])
	print '</dl>' 

def print_form( form ):
	skeys = form.keys()
	skeys.sort()
	print '<h3> The following name/value pairs ' \
	      'were entered in the form: </h3>'
	print '<dl>'
	for key in skeys:
		print '<dt>', escape(key), ':',
		print '<i>', escape(`type(form[key])`), '</i>',
		print '<dd>', escape(`form[key]`)
	print '</dl>'

def escape( s ):
	s = re.sub.gsub('&', '&amp;', s) # Must be done first
	s = re.sub.gsub('<', '&lt;', s)
	s = re.sub.gsub('>', '&gt;', s)
	return s

def test( what ):
	label = escape(str(what))
	print 'Content-type: text/html\n\n'
	print '<HEADER>\n<TITLE>' + label + '</TITLE>\n</HEADER>\n'
	print '<BODY>\n' 
	print "<H1>" + label +"</H1>\n"
	form = what()
	print_form( form )
	print_environ()
	print_environ_usage() 
	print '</body>'

if __name__ == '__main__' : 
	test_classes = ( FormContent, FormContentDict, SvFormContentDict, InterpFormContentDict )
	test( test_classes[0] )	# by default, test compatibility with 
				# old version, change index to test others.

class StrOrFileFd:
  pt = 0
  str = ''
  fd  = 0
  fileName = ''
  sourceFileName = ''
  mode = 'r'
  bufsize = -1
  strOpen = 0
  pastEnd = 0

  # Note: called again for re-opening!
  def __init__(self, fileName='', str='', mode='r', bufsize=-1, sourceFileName = ''):
    self.str = str
    self.fileName = fileName
    self.sourceFileName = sourceFileName
    self.mode = mode
    self.bufsize = bufsize
    self.pt = 0

    # Will we serve a file?
    if fileName:
      import sys
      self.fd = open(fileName, mode, bufsize)
    else:
      # Will serve a string:
      self.strOpen = 1
      self.pastEnd = len(str)
      # make mode be uniformly either r, w, or a (we don't need r+, etc):
      if string.find(mode, 'r') > -1:
	self.mode = 'r'
      elif string.find(mode, 'w') > -1:
	self.mode = 'w'
      elif string.find(mode, 'a') > -1:
	self.mode = 'a'
	self.pt = self.pastEnd
      else:
	self.pt = 'r'
    
  def reopen(self, fileName='', str='', mode='r', bufsize=-1, sourceFileName = ''):
    self.__init__(fileName=fileName,
		  str=str,
		  mode=mode,
		  bufsize=bufsize,
		  sourceFileName=sourceFileName)

  def checkStrClosed(self):
    if not self.strOpen:
      raise ValueError, 'I/O operation on closed file'

  def close(self):
    if self.fileName:
      self.fd.close()
    else:
      self.strOpen = 0
      
  def flush(self):
    if self.fileName:
      self.fd.flush()
    else:
      self.checkStrClosed()

  def isatty(self):
    if self.fileName:
      return(self.fd.isatty())
    else:
      self.checkStrClosed()
      return(0)

  def read(self, size=-1):
    if self.fileName:
      if size == -1:
        return(self.fd.read())
      else:
	return(self.fd.read(size))
    else:
      self.checkStrClosed()
      curPt = self.pt
      if self.pt >= self.pastEnd:
	return('')
      if size < -1:
	self.pt = self.pastEnd
	return(self.str[curPt:])
      else:
	self.pt = self.pt + size
	if self.pt > self.pastEnd:
	  self.pt = self.pastEnd
	  return(self.str[curPt:self.pt])
        else:
          return(self.str[curPt:curPt+size])

  def readline(self):
    if self.fileName:
      return(self.fd.readline())
    else:
      self.checkStrClosed()
      if self.pt >= self.pastEnd:
	return('')
      curPt = self.pt
      endPos = string.find(self.str[self.pt:], '\n')
      if endPos == -1:
	self.pt = self.pastEnd
	return(self.str[curPt:])
      else:
	self.pt = self.pt + endPos + 1
	return(self.str[curPt:curPt + endPos + 1])

  def readlines(self):
    if self.fileName:
      return(self.fd.readlines())
    else:
      self.checkStrClosed()
      if self.pt >= self.pastEnd:
	return('')
      res = []
      while self.pt < self.pastEnd:
	res.append(self.readline())
      return(res)

  def seek(self, offset, whence=0):
    if self.fileName:
      self.fd.seek(offset, whence)
    else:
      self.checkStrClosed()
      if whence == 0:
	self.pt = offset
      elif whence == 1:
	self.pt = self.pt + offset
      elif whence == 2:
	self.pt = (self.pastEnd) + offset
      if self.pt > self.pastEnd:
	self.pt = self.pastEnd
      elif self.pt < 0:
	self.pt = 0

  def tell(self):
    if self.fileName:
      return(self.fd.tell())
    else:
      self.checkStrClosed()
      return(self.pt)

  def write(self, theStr):
    if self.fileName:
      self.fd.write(theStr)
    else:
      if type(theStr) <> types.StringType:
	raise ValueError, "Argument to 'write' must be a string: " + str(theStr)
      self.checkStrClosed()
      if (self.mode <> 'w') and\
	 (self.mode <> 'a'):
	raise IOError, [9, 'Bad file number']
      if self.mode == 'a':
	self.str = self.str + theStr
      else:
	# Need to overwrite at pt:
	self.str = self.str[:self.pt] + theStr + self.str[self.pt + len(theStr):]
      self.pastEnd = len(self.str)
      self.pt = self.pt + len(theStr)
 	
  def writelines(self, list):
    if self.fileName:
      self.fd.writelines(list)
    else:
      self.checkStrClosed()
      for line in list:
	self.write(line)

  def name(self):
    if self.fileName:
      return(self.fd.name)
    else:
      return(self.fileName)

  def getSourceFileName(self):
    return(self.sourceFileName)






