################################################################################
#
# File:         bibcompile.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibcompile.py,v 1.3 2007/07/27 02:37:10 paepcke Exp $
# Description:  Compiling Arrays of Tagged Strings Into Documents
# Author:       Andreas Paepcke, Stanford University
# Created:      Mon Oct  9 07:59:45 1995
# Modified:     Thu Jul 26 19:35:18 2007 (Andreas Paepcke) paepcke@EMIL
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Thu Jul 26 19:35:01 2007 (Andreas Paepcke) paepcke@EMIL
#  Changed background color of output screens.
# Thu Jul 26 12:54:58 2007 (Andreas Paepcke) paepcke@EMIL
#  Started to make xhtml compliant
# Sun Jul 22 07:44:22 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Sun May  5 14:08:28 1996 (Andreas Paepcke) paepcke@AndreasHomePC.Stanford.EDU
#  Took out \n's from RTF compiler. Also escaped all backslashes. They worked
#  unescaped for some chars, but for ones that are Python special chars,
#  it didn't work.
# Mon Jan  1 12:31:05 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Made error and warning reports use bibutils.report()
# Tue Dec 19 16:18:57 1995 (Andreas Paepcke) paepcke@Stingray
#  Took out the CRs between each reference in MIF
# Thu Dec 14 11:17:41 1995 (Andreas Paepcke) paepcke@Stingray
#  Made sure that links only got inserted when the link type matched the
#  compiler system (i.e. www links would not be inserted into mif docs as
#  mif links.
################################################################################

# Public Classes and Methods:
#
#   BibHTMLCompiler
#   BibMIFCompiler
#     attr: errors
#     attr: warnings
#     showErrors
#     showWarnings
#     compileRef
#     compileRefList
#   BibRTFCompiler

# Documentation:
#
#  - New compiler: build a subclass and provide standard methods.

import string
import types
import bibrefgen
import bibconvert
import bibutils

#####################################################
#
# Exceptions
#
#####################################################

class NotSupportedYet(bibutils.InterbibException):
  pass

#####################################################
#
# Constants
#
#####################################################

MIFrefParFormat = "   <Pgf\n\
    <PgfAlignment Left >\n\
    <PgfFIndent  0.0\">\n\
    <PgfLIndent  0.3\">\n\
    <TabStop \n\
     <TSX  0.3\">\n\
     <TSType Left >\n\
     <TSLeaderStr ` '>\n\
    > # end of TabStop\n\
  > # end of Pgf (local par format)\n"

#####################################################
#
# Class Definitions
#
#####################################################

class BibCompiler:
  res = ""
  tagDispatchTable = {}  # Initialized in init method
  errors = []
  warnings = []
  verbosity = 1
  chrConverter  = bibconvert.StringPlatformConverter()

  def __init__(self):

    self.errors = []
    self.warnings = []

    self.tagDispatchTable = {
      bibrefgen.plain:[self.handlePlainOpen, self.handlePlainClose],
      bibrefgen.italic:[self.handleItalicOpen, self.handleItalicClose],
      bibrefgen.bold:[self.handleBoldOpen, self.handleBoldClose]
      }

  def showErrors(self):
    if self.errors:
      for error in self.errors:
        bibutils.report(error)
    else:
      bibutils.report("No errors.")

  def showWarnings(self):
    if self.warnings:
      for warn in self.warnings:
        bibutils.report(warn)
    else:
      bibutils.report("No warnings.")

  # Given a list of application/platform-independent reference objects, produce
  # a string that corresponds to a complete translation of the list of refs to
  # some target environment.
  # This is an abstract class!! Clients instantiate children to get the compiler
  # for their desired target. The optional 'title' is used in a compiler-specific
  # way. For example: the HTML compiler uses it to produce a web page title.

  def compileRefList(self, refList, title=""):

    self.errors = []
    self.warnings = []
    self.finalRes = ''

    # one-time opening info for the entire translation:
    self.finalRes = self.finalRes + self.opening(title)
    for ref in refList:
      self.finalRes = self.finalRes + self.compileRef(ref)
    self.finalRes = self.finalRes + self.closing()
    return(self.finalRes)
    

  # A ref is an obj containing TaggedStr's. Produce a string that will
  # successfully display the information in a specific target environment.
  # This is an abstract class!! Clients instantiate children to get the compiler
  # for their desired target.

  def compileRef(self, ref):

    self.errors = []
    self.warnings = []

    self.res = ''
    if not ref:
      return(0)
    if not bibutils.isInstanceOf(ref, bibrefgen.BibReference):
      self.errors.append("Error: Arg to 'compile' method must be list of ref objs.")
      if self.verbosity:
	raise TypeError, "Arg to 'compile' method must be list of ref objs."
      else:
	return(0)

    self.handleRefOpen()
    for tstr in ref.refSpec:
      try:
        for tag in tstr.tag:
	  if type(tag) == types.InstanceType:
	    self.handleLinkOpen(tag)
	  else:
	    tagOpenMethod = self.tagDispatchTable[tag][0]
	    tagOpenMethod()
	self.handleStr(tstr.str, ref)
        for tag in tstr.tag:
	  if type(tag) == types.InstanceType:
	    self.handleLinkClose(tag)
	  else:
	    tagCloseMethod = self.tagDispatchTable[tag][1]
            tagCloseMethod()
      except KeyError:
	self.errors.append("Unknown tag: '"\
			   + str(tstr.tag)\
			   + "'. Substring ignored!")
    self.handleRefClose()
    return(self.res)   

  def opening(self, title=""):
    return('')

  def closing(self):
    return('')

  def handleStr(self, str, refObj):
    self.res = self.res + str

  def handleRefOpen(self):
    pass

  def handleRefClose(self):
    pass


class BibHTMLCompiler(BibCompiler):

  def handleStr(self, str, refObj):
    if refObj.charSet == bibconvert.Latin1CharSet:
      self.res = self.res + str
    else:
      self.res = self.res +\
		 self.chrConverter.cnvSetToSet(str,
					       refObj.charSet,
					       bibconvert.Latin1CharSet)

  def opening(self, title=''):
    if title:
      return('<html><head><TITLE>' + title + '</TITLE></head><body bgcolor="#FFECD9">\n' +\
	     '<center><H1>' + title + '</H1></center>\n<p></p>\n')
    else:
      return('<html><head><TITLE>' + title + '</TITLE></head><body bgcolor="#FFECD9">\n')

  def closing(self):
    return('</body></html>\n')


  def handleRefOpen(self):
    self.res = self.res + '<p></p>\n'


  def handlePlainOpen(self):
    pass

  def handleItalicOpen(self):
    self.res = self.res + '<i>'
      
  def handleBoldOpen(self):
    self.res = self.res + '<b>'
      
  def handlePlainClose(self):
    pass

  def handleItalicClose(self):
    self.res = self.res + '</i>'
      
  def handleBoldClose(self):
    self.res = self.res + '</b>'

  def handleLinkOpen(self, linkObj):
    if linkObj.system == 'www':
      self.res = self.res + '<a href="' + linkObj.destination + '">'

  def handleLinkClose(self, linkObj):
    if linkObj.system == 'www':
      self.res = self.res + '</a>'

class BibMIFCompiler(BibCompiler):

  refStart    = "<Para\n" + MIFrefParFormat + "  <ParaLine \n"
  #refEnd      = "  > # end ParaLine\n> # end of para\n<Para <ParaLine>>\n"
  refEnd      = "  > # end ParaLine\n> # end of para\n"
  italicStart = "    <Font <FTag `'> <FAngle `Italic'> >\n"
  boldStart   = "    <Font <FTag `'> <FWeight `Bold'> >\n"
  italicEnd   = "    <Font <FTag `'> >\n"
  boldEnd     = "    <Font <FTag `'> >\n"
  strStart    = "    <String `"
  strEnd      = "'>\n"
  linkStart   = "    <Marker\n      <MType 8>\n      <MText `"
  linkEnd     = "    >\n"


  def handleStr(self, str, refObj):

    self.res = self.res + self.strStart
    if refObj.charSet == bibconvert.MIFCharSet:
      self.res = self.res + str
    else:
      self.res = self.res +\
		 self.chrConverter.cnvSetToSet(str,
					       refObj.charSet,
					       bibconvert.MIFCharSet)
    self.res = self.res + self.strEnd

  def opening(self, title=''):
    return("<MIFFile 5.00> # Generated by InterBib\n")

  def closing(self):
    return("\n# end of bibliography")

  def handleRefOpen(self):
    self.res = self.res + self.refStart

  def handleRefClose(self):
    self.res = self.res + self.refEnd

  def handlePlainOpen(self):
    pass

  def handleItalicOpen(self):
    self.res = self.res + self.italicStart
      
  def handleBoldOpen(self):
    self.res = self.res + self.boldStart
      
  def handlePlainClose(self):
    pass

  def handleItalicClose(self):
    self.res = self.res + self.italicEnd
      
  def handleBoldClose(self):
    self.res = self.res + self.boldEnd


  def handleLinkOpen(self, linkObj):
    if linkObj.system == 'framemaker':
      self.res = self.res + self.linkStart + linkObj.destination + "'>\n"

  def handleLinkClose(self, linkObj):
    if linkObj.system == 'framemaker':
      self.res = self.res + self.linkEnd



class BibRTFCompiler(BibCompiler):

  def handleStr(self, str, refObj):
    if refObj.charSet == bibconvert.RTFCharSet:
      self.res = self.res + str
    else:
      self.res = self.res +\
        self.chrConverter.cnvSetToSet(str,
                                      refObj.charSet,
                                      bibconvert.RTFCharSet)

  def opening(self, title=''):
    if title:
      return "{\\rtf1\\ansi{\\info{\\title " + title + "}} \\pard\\plain " +\
             "\\qc {\\b\\fs24 " + title + "}\n\\par\\pard"
    else:
      return "{\\rtf1\\ansi "

  def closing(self):
    return "}"

  def handleRefOpen(self):
    self.res = self.res + "\\pard\\plain \\par "

  def handleRefClose(self):
    self.res = self.res + "\\par "

  def handlePlainOpen(self):
    pass

  def handleItalicOpen(self):
    self.res = self.res + "{\\i "
      
  def handleBoldOpen(self):
    self.res = self.res + "{\\b "
      
  def handlePlainClose(self):
    pass

  def handleItalicClose(self):
    self.res = self.res + "}"
      
  def handleBoldClose(self):
    self.res = self.res + "}"

#
# The following are empty until I can find a linking facility in RTF
#

  def handleLinkOpen(self, linkObj):
    pass
#   if linkObj.system == 'framemaker':
#     self.res = self.res + self.linkStart + linkObj.destination + "'>\n"

  def handleLinkClose(self, linkObj):
    pass
#   if linkObj.system == 'framemaker':
#     self.res = self.res + self.linkEnd
