################################################################################
#
# File:         bibconvert.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibconvert.py,v 1.10 2007/08/09 02:42:18 paepcke Exp $
# Description:  Character set converters
# Author:       Andreas Paepcke, Stanford University
# Created:      Wed Oct 11 11:26:30 1995
# Modified:     Wed Aug 08 19:09:36 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Mon Aug 06 20:00:02 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Slimmed down the Refer dictionaries to adhere to stricter
#  standard.
# Thu Aug 02 10:57:52 2007 (Andreas Paepcke) paepcke@EMIL
#  Changed class RecordTypeNameConverterEndnoteInterbib to
#  RecordTypeNameConverterBibtexInterbib.
# Thu Aug 02 10:53:33 2007 (Andreas Paepcke) paepcke@EMIL
#  Added Endnote Attribute Converter class.
# Tue Jul 31 07:46:34 2007 (Andreas Paepcke) paepcke@EMIL
#  Changed %R to map to url field, rather than title.
# Mon Jul 30 18:36:41 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  In extractLatexIgnorableControlStr(): added \url
# Sun Jul 29 12:50:04 2007 (Andreas Paepcke) paepcke@EMIL
#  Modified the Refer translation table to map entryType-->0, key-->F,
#  and label-->L. Analogously to the reverse table. Endnote stores
#  'labels' in %F. That is one field that's imported for all types
#  of pubs (proceedings, books, etc.). So we use it to store our
#  BibTeX keys there. Endnote expects in %0 the record type (conference proceedings, etc.)
# Sun Jul 22 07:47:34 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Fri May  8 12:43:18 1998 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Allowed previously untranslatable control sequences (as listed in
#  nonTranslatableLatexControls). Made it so that the corresponding
#  chars are used without the modifyer instead. As in: \v{o} used simply
#  as a normal 'o'.
# Fri May  8 10:48:23 1998 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Allowed dumb Latex cnt sequences like: "\'97". Turning into a straight
#  apostrophe now.
# Thu May  7 14:06:11 1998 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added handling of Latex cnt seqs: '\v{x}' and '{\v x}' which would
#  normally put a 'v' on top of the operand letter ('x' in above ex.). Since
#  the Frame char set doesn't support this, we ignore it and use just 'x'.
# Mon Mar 31 18:35:29 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added translation for BibTeX idiom \~{} which is used to create a ~ char.
# Thu Sep 19 08:31:06 1996 (Andreas Paepcke) paepcke@Walrus
#  Fixed cases like: "This looks like \ {T}rouble". System saw the opening
#  '{' and thought it would get something like '\^{a}'.
# Mon Jul  1 19:29:42 1996 (Andreas Paepcke) paepcke@Stingray
#  Moved import of USMarc down into the file to avoid cross-import problem
#  with USMarc.py.
# Fri Jun 14 14:28:27 1996 (Andreas Paepcke) paepcke@Stingray
#  In raw dl converter: changed to assume that conversions are from USMarc to
#  local attributes. 
# Sun May  5 10:19:35 1996 (Andreas Paepcke) paepcke@AndreasHomePC.Stanford.EDU
#  Added conversion support for RTF.
# Tue Feb  6 22:21:50 1996 (Andreas Paepcke) paepcke@Stingray
#  Added support for special char shortcuts, such as "\^o" or "\'e"
# Sat Feb  3 22:47:02 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Changed Refer attribute translation of %X to be 'annote', instead of
#  'abstract'. This way it gets into annotated bibliographies.
# Tue Dec 26 15:22:46 1995 (Andreas Paepcke) paepcke@Stingray
#  In frameToLatin1: made sure that *both* the mapping from
#  dagger *and* double dagger to space are taken out of 
#  specialLatin1Chars.
# Tue Dec 26 14:58:46 1995 (Andreas Paepcke) paepcke@Stingray
#  cnvLatin1ToFrame was accidentally overwritten by what was
#  intended to be cnvMIFToFrame!!! Fixed that and caught error
#  in cnvMIFToFrame when cntrol seq was not found.
# Mon Dec 11 16:31:59 1995 (Andreas Paepcke) paepcke@Stingray
#  Added FrameTAB constant and resolved circular bibconvert/bibdb load.
# Tue Dec  5 08:41:01 1995 (Andreas Paepcke) paepcke@Stingray
#  Added ReferAttributeConverter
# Tue Oct 24 18:46:14 1995 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Added DialogAttributeConverter, FolioAttributeConverter,
#    WebAttributeConverter. Made RawDLDocumentAttributeConverter. They all
#    inherit from AttributeConverter
################################################################################

# To do:
#  - macintosh char set

import string
import bibutils

# The following import needs to happen later to avoid conflict with USMarc loading:
#import USMarc


#####################################################
#
# Public Constants
#
#####################################################

FrameCharSet = 'FrameCharSet'
MIFCharSet   = 'MIFCharSet'
RTFCharSet   = 'RTFCharSet'
LatexCharSet = 'LatexCharSet'
Latin1CharSet = 'Latin1CharSet'
MacintoshCharSet = 'MacintoshCharSet'

legalCharSets = [FrameCharSet, MIFCharSet, LatexCharSet,
		 Latin1CharSet, RTFCharSet]

FrameTAB = '\x08'
RTFTAB = FrameTAB

# Reserved name of the field for links.
# NOTE: If you change it here, also change it in bibutils!
#       There should be a bibconstants.py to have all these
#       constants. But there isn't right now.
# We have a cross-import problem between this
# bibconvert.py and the bibutils.py file. If you change
# the constant here, change in as well in bibutils.py!!

linksFieldName = 'links'

# BibTeX field name for a URL:

urlFieldName   = 'url'

# Name of field in entries for the BibTeX record key.
# This is the info right after the @...{

keyFieldName = 'entryKey'

# Name of field in entries that has the 
# type of record as a string ("inproceedings" etc.):

typeFieldName = 'entryType'


#####################################################
#
# Exceptions
#
#####################################################

class UnsupportedCharSet(bibutils.InterbibException):
  pass

class UnknownOrNonTranslateableLatexControlSeq(bibutils.InterbibException):
  pass

class NonTranslatableAttribute(bibutils.InterbibException):
  pass


#####################################################
#
# Translation Tables Used By BibEntry Class
#
#####################################################

# These are taken in order from FrameMaker 5 'Quick Reference Manual':
frameToLatex = {
  '\x23': '#'   , '\x24': '$'   , '\x25': '%'   ,
  '\x26': '&'   , '\x40': '@'   , '\x5f': '_'   ,
  '\x7b': '{'   , '\x7d': '}'   , '\x80': '"{A}',
  '\x81': 'AA'  , '\x82': 'c{C}', '\x83': "'{E}", '\x84': '~{N}',
  '\x85': '"{O}', '\x86': '"{U}', '\x87': "'{a}", '\x88': '`{a}',
  '\x89': '^{a}', '\x8a': '"{a}', '\x8b': '~{a}', '\x8c': 'aa'  ,
  '\x8d': 'c{c}', '\x8e': "'{e}", '\x8f': '`{e}', '\x90': '^{e}',
  '\x91': '"{e}', '\x92': "'{i}", '\x93': '`{i}', '\x94': '^{i}',
  '\x95': '"{i}', '\x96': '~{n}', '\x97': "'{o}", '\x98': '`{o}',
  '\x99': '^{o}', '\x9a': '"{o}', '\x9b': '~{o}', '\x9c': "'{u}",
  '\x9d': '`{u}', '\x9e': '^{u}', '\x9f': '"{u}', '\xa0': 'dag' ,
  '\xbe': 'ae'  , '\xbf': 'o'   , '\xcb': '`{A}', '\xcc': '~{A}',
  '\xcd': '~{O}', '\xce': 'OE'  , '\xcf': 'oe'  , '\xd0': '--'  ,
  '\xd1': '---' , '\xe0': 'ddag', '\xe5': '^{A}', '\xe6': '^{E}',
  '\xe7': "'{A}", '\xe8': '"{E}', '\xe9': '`{E}', '\xea': "'{I}",
  '\xeb': '^{I}', '\xec': '"{I}', '\xed': '`{I}', '\xee': "'{O}",
  '\xef': '^{O}', '\xf1': '`{O}', '\xf2': "'{U}", '\xf3': '^{U}',
  '\xf4': '`{U}', '\x7e': '~{}'
  }

# Latex chars that have no equivalence in the Frame set and will be
# replaced with the regular chars. As in: \v{o} which would put a 'v'
# on top of the 'o'. This will just come out as 'o':
nonTranslatableLatexControls = ['.','u','v','H','t','d','b']
specialFrameLatexChars = frameToLatex.keys()	   
frameEnDash = '\xd0'
frameEmDash = '\xd1'

frameToLatin1 = {
  '\x80': '\xc4', '\x81': '\xc5', '\x82': '\xc7', '\x83': '\xc9',
  '\x84': '\xd1', '\x85': '\xd6', '\x86': '\xdc', '\x87': '\xe1',
  '\x88': '\xe0', '\x89': '\xe2', '\x8a': '\xe4', '\x8b': '\xe3',
  '\x8c': '\xe5', '\x8d': '\xe7', '\x8e': '\xe9', '\x8f': '\xe8',
  '\x90': '\xea', '\x91': '\xeb', '\x92': '\xed', '\x93': '\xec',
  '\x94': '\xee', '\x95': '\xef', '\x96': '\xf1', '\x97': '\xf3',
  '\x98': '\xf2', '\x99': '\xf4', '\x9a': '\xf6', '\x9b': '\xf5',
  '\x9c': '\xfa', '\x9d': '\xf9', '\x9e': '\xfb', '\x9f': '\xfc',
  '\xa0': ' '   , '\xbe': '\xe6', '\xbf': '\xf8', '\xcb': '\xc0',
  '\xcc': '\xc3', '\xcd': '\xd5', '\xce': 'OE'  , '\xcf': 'oe'  ,
  '\xd0': '-'   , '\xd1': '-'   , '\xe0': ' '   , '\xe5': '\xc2',
  '\xe6': '\xca', '\xe7': '\xc1', '\xe8': '\xcb', '\xe9': '\xc8',
  '\xea': '\xc9', '\xeb': '\xce', '\xec': '\xcf', '\xed': '\xcc',
  '\xee': '\xd3', '\xef': '\xd4', '\xf1': '\xd2', '\xf2': '\xda',
  '\xf3': '\xdb', '\xf4': '\xd9'
  }

specialFrameLatin1Chars = frameToLatin1.keys()	   
specialLatin1Chars = frameToLatin1.values()	   

# Delete the chars we are approximating to (' ' (for dagger and double
# dagger), 'OE', 'oe') so that they will be left alone when xlating from
# Latin1 to Frame:

specialLatin1Chars.remove(' ') # for double dagger
specialLatin1Chars.remove(' ') # for double dagger
specialLatin1Chars.remove('-')
specialLatin1Chars.remove('OE')
specialLatin1Chars.remove('oe')

# MIF needs escapes for tab, >, ', and \. Note that like in Latex,
# we don't include the backslashes in the table:

frameToMIF = {
  '\x08': 't', '\x3e': '>', '\x27': 'q', '\x60': 'Q', '\x5c': '\\'
  }

specialFrameMIFChars = frameToMIF.keys()	   
specialMIFChars = frameToMIF.values()	   

# RTF needs escapes for {, }, \, `, ', ", -, --Note that like in Latex,
# we don't include the backslashes in the table. We also do xlations for
# Frame's left and right double-quotes.

frameToRTF = {
  FrameTAB: 'tab ',
  '\x5c': '\\' ,   '\x7b': '{'  ,   '\x7d': '}'  ,

  '\x80': "{\\cchs0 \\'c4}",   '\x81': "{\\cchs0 \\'c5}",   '\x82': "{\\cchs0 \\'c7}",   '\x83': "{\\cchs0 \\'c9}",
  '\x84': "{\\cchs0 \\'d1}",   '\x85': "{\\cchs0 \\'d6}",   '\x86': "{\\cchs0 \\'dc}",   '\x87': "{\\cchs0 \\'8e}",
  '\x88': "{\\cchs0 \\'e0}",   '\x89': "{\\cchs0 \\'e2}",   '\x8a': "{\\cchs0 \\'e4}",   '\x8b': "{\\cchs0 \\'e3}",
  '\x8c': "{\\cchs0 \\'e5}",   '\x8d': "{\\cchs0 \\'e7}",   '\x8e': "{\\cchs0 \\'e9}",   '\x8f': "{\\cchs0 \\'e8}",
  '\x90': "{\\cchs0 \\'ea}",   '\x91': "{\\cchs0 \\'eb}",   '\x92': "{\\cchs0 \\'ed}",   '\x93': "{\\cchs0 \\'ec}",
  '\x94': "{\\cchs0 \\'ee}",   '\x95': "{\\cchs0 \\'ef}",   '\x96': "{\\cchs0 \\'f1}",   '\x97': "{\\cchs0 \\'f3}",
  '\x98': "{\\cchs0 \\'f2}",   '\x99': "{\\cchs0 \\'f4}",   '\x9a': "{\\cchs0 \\'f6}",   '\x9b': "{\\cchs0 \\'f5}",
  '\x9c': "{\\cchs0 \\'fa}",   '\x9d': "{\\cchs0 \\'f9}",   '\x9e': "{\\cchs0 \\'fb}",   '\x9f': "{\\cchs0 \\'fc}",
  '\xa0': "{\\cchs0 \\'86}",   '\xa7': "{\\cchs0 \\'df}",   '\xae': "{\\cchs0 \\'c6}",   '\xaf': "{\\cchs0 \\'d8}",
  '\xbe': "{\\cchs0 \\'e6}",   '\xbf': "{\\cchs0 \\'f8}",   '\xcb': "{\\cchs0 \\'c0}",   '\xcc': "{\\cchs0 \\'c3}",
  '\xcd': "{\\cchs0 \\'d5}",   '\xce': "{\\cchs0 \\'8c}",   '\xcf': "{\\cchs0 \\'9c}",

  '\xd0': "endash ", '\xd1': "emdash ", '\xd2': "ldblquote ", '\xd3': "rdblquote ",
  '\xd4': 'lquote ', '\xd5': 'rquote ', '\xd8': "{\\cchs0 \\'ff}",

  '\xe0': "{\\cchs0 \\'87}",   '\xe5': "{\\cchs0 \\'c2}",   '\xe6': "{\\cchs0 \\'ca}",   '\xe7': "{\\cchs0 \\'c1}",
  '\xe8': "{\\cchs0 \\'cb}",   '\xe9': "{\\cchs0 \\'c8}",   '\xea': "{\\cchs0 \\'cd}",   '\xeb': "{\\cchs0 \\'ce}",
  '\xec': "{\\cchs0 \\'cf}",   '\xed': "{\\cchs0 \\'cc}",   '\xee': "{\\cchs0 \\'d3}",   '\xef': "{\\cchs0 \\'d4}",
  '\xf1': "{\\cchs0 \\'d2}",   '\xf2': "{\\cchs0 \\'da}",   '\xf3': "{\\cchs0 \\'db}",   '\xf4': "{\\cchs0 \\'d9"
  }   

specialFrameRTFChars = frameToRTF.keys()	   
specialRTFChars = frameToRTF.values()	   

# Needs filling in ***
# Dictionary from FrameMaker special chars to Macintosh special chars:
frameToMac = {
  '\x80': '\212','\x81': '\214', '\x82': '\215'}

# Take the frameToLatex dictionary and build the inverse:
latexToFrame = {}
for key in frameToLatex.keys():
  latexToFrame[frameToLatex[key]] = key
# All the various Latex spaces are translated to regular spaces (\@, etc):
latexToFrame[','] = '\x20'
latexToFrame[':'] = '\x20'
latexToFrame[';'] = '\x20'
latexToFrame[' '] = '\x20'
latexToFrame['@'] = '\x20'
latexToFrame['/'] = '\x20'
# Ignore the hyphen advisory '\-':
latexToFrame['-'] = ''
latexToFrame['\\'] = '\x5c'
latexToFrame['par'] = '\x0a'
# The 'i' with cedi is made into a regular 'i':
#latexToFrame['c{i}'] = '\x69'
# Take care of inner back-slashed controls like: '\"{\i}' (the dot-less 'i'. We just
# translate this as a dotted 'i':
latexToFrame["'{\\i}"] = latexToFrame["'{i}"]
latexToFrame["`{\\i}"] = latexToFrame["`{i}"]
latexToFrame["^{\\i}"] = latexToFrame["^{i}"]
latexToFrame['"{\\i}'] = latexToFrame['"{i}']

# Same for macToFrame:
macToFrame = {}
for key in frameToMac.keys():
  macToFrame[frameToMac[key]] = key
specialMacChars = macToFrame.keys()

# And for latin1ToFrame:
latin1ToFrame = {}
for key in frameToLatin1.keys():
  latin1ToFrame[frameToLatin1[key]] = key

# And for MIFToFrame:
MIFToFrame = {}
for key in frameToMIF.keys():
  MIFToFrame[frameToMIF[key]] = key

# And for RTFToFrame:
RTFToFrame = {}
for key in frameToRTF.keys():
  RTFToFrame[frameToRTF[key]] = key

# Now that the inverse of frameToRTF is computed, we add some xlations:
# The Frame straight double-quotes (as opposed to left/right smart quotes)
# are treated like left double quotes. The single straight quote is treated
# like a left quote:

frameToRTF['\x22'] = 'ldblquote '
frameToRTF['\x27'] = 'lquote '
specialFrameRTFChars = specialFrameRTFChars + ['\x22', '\x27']

class StringPlatformConverter:
  errors = []
  warnings = []

  def __init__(self):
    errors = []
    warnings = []

  def showErrors(self):
    if self.errors:
      for error in self.errors:
        print error
    else:
      print "No errors."

  def showWarnings(self):
    if self.warnings:
      for warn in self.warnings:
        print warn
    else:
      print "No warnings."

  # Given two char sets and a string, assume that the string is in the
  # first set. Convert it to the second:

  def cnvSetToSet(self, str, fromSet, toSet):

    self.errors = []
    self.warnings = []

    if fromSet not in legalCharSets:
      raise UnsupportedCharSet, fromSet
    if toSet not in legalCharSets:
      raise UnsupportedCharSet, toSet

    if fromSet == toSet:
      return(str)
    elif fromSet <> FrameCharSet:
      # Convert to Frame char set as intermediate
      if fromSet == LatexCharSet:
	str = self.cnvLatexToFrame(str)
      elif fromSet == MacintoshCharSet:
	str = self.cnvMacintoshToFrame(str)
      elif fromSet == Latin1CharSet:
	str = self.cnvLatin1ToFrame(str)
      elif fromSet == MIFCharSet:
	str = self.cnvMIFToFrame(str)
      elif fromSet == RTFCharSet:
	str = self.cnvRTFToFrame(str)
    # Now we have Framemaker char set. Convert to final target:
    if toSet == FrameCharSet:
      return(str)
    elif toSet == LatexCharSet:
      return(self.cnvFrameToLatex(str))
    elif toSet == MacintoshCharSet:
      return(self.cnvFrameToMacintosh(str))
    elif toSet == Latin1CharSet:
      return(self.cnvFrameToLatin1(str))
    elif toSet == MIFCharSet:
      return(self.cnvFrameToMIF(str))
    elif toSet == RTFCharSet:
      return(self.cnvFrameToRTF(str))

  # Given a string (potentially) with FrameMaker special chars, produce Latex
  # output:

  def cnvFrameToLatex(self, str):

    self.errors = []
    self.warnings = []
    convStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      if chr in specialFrameLatexChars:
        #Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char with a leading backslash, unless
	# it's an endash or emdash. They don't get backslashes:
	if (chr == frameEnDash) or (chr == frameEmDash):
	  convStr = convStr +\
		    str[substrIndx:i] +\
		    frameToLatex[chr]
	else:
	  convStr = convStr +\
		    str[substrIndx:i] +\
		    '\\' +\
		    frameToLatex[chr]
        substrIndx = i + 1
    return(convStr + str[substrIndx:])

  # Given a string (potentially) including Latex special chars, convert
  # them to FrameMaker. This includes things like '\"{a}' but also just
  # curly for forcing capitalization. Those are deleted. We go through the
  # str looking for control chars and produce a new output str and we go.
  # It has the control chars replaced with Framemaker chars. For
  # efficiency, we don't just copy each char. Instead be keep two pts: i is
  # the pt going through str. substrIndx is also a pt into str that trails
  # behind i. We only copy pieces from str to the output string when we
  # reach a cnt char. If there are no control chars, this means that the
  # whole string is just copied over at the end.

  def cnvLatexToFrame(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    cntrStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      # Ignore entry into math mode because we don't support it:
      if (chr == "$") and i >= substrIndx:
        self.warnings.append("Ignoring LaTeX math mode entry or exit")
	convStr = convStr + str[substrIndx:i]
	substrIndx = i + 1
	continue
      # Replace '~' with a regular space
      if (chr == "~") and i >= substrIndx:
	convStr = convStr + str[substrIndx:i] + ' '
	substrIndx = i + 1
	continue
      # Delete curly braces that are used to force capitalization:
      if (chr == "{" or chr == "}") and i >= substrIndx:
	convStr = convStr + str[substrIndx:i]
	substrIndx = i + 1
	continue
      # EM-Dash or EN-Dash?
      if chr == "-":
        if (string.find(str, "---", i) == i) and i >= substrIndx:
  	  convStr = convStr + str[substrIndx:i] + latexToFrame['---']
	  substrIndx = i + 3
	  continue
	elif (string.find(str, "--", i) == i) and i >= substrIndx:
  	  convStr = convStr + str[substrIndx:i] + latexToFrame['--']
	  substrIndx = i + 2
	  continue
      # Beginning of a control sequence (e.g. "\'{a}" or "\aa")?
      if chr == '\\' and i >= substrIndx:
	wasSpecialAbbreviation = 0
	ignorable = self.extractLatexIgnorableControlStr(str, i)
	if (len(str) > i+2) and\
	   str[i+2] == "{" and\
	   not self.extractLatexControlStr(str, i):
	  # It's of the form "\'{a}" or "\t{oo}" or "\'{\i}":
          # Special trouble: Framemaker does not support some cntrl chrs
	  # that are legal in Latex. Ex: {\v{o}}. So we ignore those modifiers
	  # the '\v' (or \H or whatever) part. The curly braces will be
	  # taken care of above:
	  if str[i+1] in nonTranslatableLatexControls:
	    if len(str) > i+5:
              self.warnings.append("Cannot translate special char '" + str[i:i+5] + "'. Using char without modifyer instead.")
	    else:
	      self.warnings.append("Cannot translate special char '\\" + str([i+1]) + "'. Using char without modifyer instead.")
	    convStr = convStr + str[substrIndx:i]
	    # Skip the '\x':
   	    substrIndx = i + 2
	    continue
	  # It's a managable control seq:
	  seqEnd = string.find(str, "}", i+3)
	  cntrStr = str[i+1:seqEnd+1]
        # Check for control sequences that are shortcuts of the above:
	# like "\'o" or "\^e":
	elif (len(str) > i) and self.isSpecialLatexCharShortcut(str, i) and not ignorable:
	  # Replace the cntr seq w/ the full curly-delimited one for
	  # translation purposes:
	  cntrStr = str[i+1:i+2] + '{' + str[i+2:i+3] + '}'
	  wasSpecialAbbreviation = 1
	# Check for Latex controls \xxxx that get ignored:
	elif ignorable:
	  # Ignore the control sequence:
  	  convStr = convStr + str[substrIndx:i]
	  # Move past the control seq "\it", etc and past the space after it:
	  substrIndx = i + len(ignorable) + 1
	  # Don't give warnings for every ignored contr seq like "\it", etc.
	  # But user should see when we ignored something like '{\v e}' which
	  # we can't translate:
	  if (len(ignorable) > 1) and ignorable[1] in nonTranslatableLatexControls:
            self.warnings.append("Cannot translate special char '" + ignorable + "'. Using char without modifyer instead.")
	  continue
	else:
	  # It's of the form "\aa" or "\o":
	  cntrStr = self.extractLatexControlStr(str, i)
        # End control sequence parsing. Now translate it.
	if not latexToFrame.has_key(cntrStr):
	  # Yet another exception: some people use "\'" indiscriminately,
	  # as in "\'97". In that case, we end up here with a control
	  # string like "'{9}" which will not translate propertly. In that
	  # case we just turn the "\'" into a regular "'" by ignoring the '\':
	  if cntrStr and (cntrStr[0] == "'"):
    	    convStr = convStr + str[substrIndx:i]
	    # Move past the '\':
	    substrIndx = i + 1
	    continue
	  # All right, finally: unknown control string:
	  if cntrStr:
	    raise UnknownOrNonTranslateableLatexControlSeq, '\\' + cntrStr
	  else:
	    # Couldn't even parse the contr string, so return some context:
	    raise UnknownOrNonTranslateableLatexControlSeq, '\\' + str[i:i+10] + '...'
	else:
          replacement = latexToFrame[cntrStr]
	# Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
	   str[substrIndx:i] +\
           replacement
	# Advance the substr index to pt past the control seq:
	substrIndx = i + 1 + len(cntrStr)
	# If this was a special-char abbrev: compensate for the added '{}':
	if wasSpecialAbbreviation:
	  substrIndx = substrIndx - 2
      # End of control sequence processing
    return(convStr + str[substrIndx:])
    
  # Assume that indx pts to a backslash char in the str. Return 1 if the
  # following char is the start of an abbreviated special char Latex control
  # sequence, such as "\'o" or "\^e":

  def isSpecialLatexCharShortcut(self, str, i):
    if str[i+1] in ["'", "`", "^", '"', "~", "c"]:
      return(1)
    else:
      return(0)
    

  # Given a string and a start index, return a string with a Latex control
  # sequence, if it starts at the index and is a control that can simply be
  # ignored. These are sequences of the form \em, \it, etc.  We assume the
  # start index points to the backslash that starts the sequence. Returns
  # empty string if nothing is found:

  def extractLatexIgnorableControlStr(self, str, i):
    if string.find(str, "rm", i+1) == i + 1 or\
       string.find(str, "em", i+1) == i + 1 or\
       string.find(str, "bf", i+1) == i + 1 or\
       string.find(str, "it", i+1) == i + 1 or\
       string.find(str, "sl", i+1) == i + 1 or\
       string.find(str, "sf", i+1) == i + 1 or\
       string.find(str, "sc", i+1) == i + 1 or\
       string.find(str, "tt", i+1) == i + 1:
      return(str[i:i+3])
    elif string.find(str, "cite", i+1) == i + 1:
      return(str[i:i+5])
    elif string.find(str, "url", i+1) == i + 1:
      return(str[i:i+4])
    # Special case: Framemaker does not support chars w/ a 'v' on top,
    # and six others. Just ignore those:
    elif str[i+1] in nonTranslatableLatexControls:
      return(str[i:i+2])
    else:
      return("")

  # Given a string and a start index, return a string with a Latex
  # control sequence, if it starts at the index. These are sequences of
  # the form \o or \AA. Not of the form \^{o}. We assume the start
  # index points to the backslash that starts the sequence:

  def extractLatexControlStr(self, str, i):
    if string.find(str, "dag", i+1) == i+1:    # dagger
      return("dag")
    elif string.find(str, "ddag", i+1) == i+1:   # double dagger
      return("ddag")
    elif string.find(str, "S", i+1) == i+1:     # Legal par sign
      return("S")
    elif string.find(str, "P", i+1) == i+1:
      return("S")
    elif string.find(str, "copyright", i+1) == i+1:
      return("copyright")
    elif string.find(str, "pounds", i+1) == i+1:
      return("pounds")
    elif string.find(str, "oe", i+1) == i+1:
      return("oe")
    elif string.find(str, "OE", i+1) == i+1:
      return("OE")
    elif string.find(str, "ae", i+1) == i+1:
      return("ae")
    elif string.find(str, "AE", i+1) == i+1:
      return("AE")
    elif string.find(str, "aa", i+1) == i+1:
      return("aa")
    elif string.find(str, "AA", i+1) == i+1:
      return("AA")
    elif string.find(str, "o", i+1) == i+1:
      return("o")
    elif string.find(str, "O", i+1) == i+1:
      return("O")
    elif string.find(str, "l", i+1) == i+1:
      return("l")
    elif string.find(str, "L", i+1) == i+1:
      return("L")
    elif string.find(str, "ss", i+1) == i+1:
      return("ss")
    elif string.find(str, "-", i+1) == i+1:
      return("-")      
    elif string.find(str, ",", i+1) == i+1:
      return(",")      
    elif string.find(str, ":", i+1) == i+1:
      return(":")      
    elif string.find(str, ";", i+1) == i+1:
      return(";")      
    elif string.find(str, " ", i+1) == i+1:
      return(" ")
    elif string.find(str, "@", i+1) == i+1:
      return("@")
    elif string.find(str, "/", i+1) == i+1:
      return("/")
    elif string.find(str, "$", i+1) == i+1:
      return("$")
    elif string.find(str, "%", i+1) == i+1:
      return("%")
    elif string.find(str, "&", i+1) == i+1:
      return("&")
    elif string.find(str, "#", i+1) == i+1:
      return("#")
    elif string.find(str, "_", i+1) == i+1:
      return("_")
    elif string.find(str, "{", i+1) == i+1:
      return("{")
    elif string.find(str, "}", i+1) == i+1:
      return("}")
    elif string.find(str, "\\", i+1) == i+1:
      return("\\")
    elif string.find(str, "par", i+1) == i+1:
      return("par")
    elif string.find(str, "TeX", i+1) == i+1:
      return("TeX")
    elif string.find(str, "LaTeX", i+1) == i+1:
      return("LaTeX")
    else:
      return("")

  # Given a string (potentially) with FrameMaker special chars, produce Latin1
  # output:

  def cnvFrameToLatin1(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      if chr in specialFrameLatin1Chars:
        #Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
		  str[substrIndx:i] +\
		  frameToLatin1[chr]
        substrIndx = i + 1
    return(convStr + str[substrIndx:])
    
  # Given a string (potentially) with Latin1 special chars, produce Frame
  # output:

  def cnvLatin1ToFrame(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      if chr in specialLatin1Chars:
        #Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
		  str[substrIndx:i] +\
		  latin1ToFrame[chr]
        substrIndx = i + 1
    return(convStr + str[substrIndx:])

  # Given a string (potentially) with FrameMaker special chars, produce MIF
  # output:

  def cnvFrameToMIF(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      if chr in specialFrameMIFChars:
        #Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
		  str[substrIndx:i] +\
		  '\\' +\
		  frameToMIF[chr]
        substrIndx = i + 1
    return(convStr + str[substrIndx:])
    
  # Given a string (potentially) with MIF special chars, produce Frame
  # output:

  def cnvMIFToFrame(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    cntrStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      # Beginning of a control sequence (e.g. "\t")?
      if chr == '\\':
	cntrStr = self.extractMIFControlStr(str, i)
        if cntrStr:	
          replacement = MIFToFrame[cntrStr]
	else:
	  replacement = ''
	# Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
	   str[substrIndx:i] +\
           replacement
	# Advance the substr index to pt past the control seq:
	substrIndx = i + 1 + len(cntrStr)
      # End of control sequence processing
    return(convStr + str[substrIndx:])


  # Given a string and a start index, return a string with a MIF control
  # sequence, if it starts at the index. These are sequences of the form \t
  # or \\. Assumes that index points to the backslash that starts the
  # sequence:

  def extractMIFControlStr(self, str, i):
    if string.find(str, "t", i+1) == i+1:
      return("t")
    elif string.find(str, ">", i+1) == i+1:
      return(">")
    elif string.find(str, "q", i+1) == i+1:
      return("q")
    elif string.find(str, "Q", i+1) == i+1:
      return("Q")
    elif string.find(str, "\\", i+1) == i+1:
      return("\\")
    else:
      return('')

  # Given a string (potentially) with FrameMaker special chars, produce RTF
  # output. Note that the frameToRTF table is a bit different from the
  # MIF equivalent. The translations starting with '{\\cchs' do not get
  # escaped with a backslash.

  def cnvFrameToRTF(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      if chr in specialFrameRTFChars:
	xlation = frameToRTF[chr]
        #Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	if string.find(xlation, '{\\cchs') > -1:
	  convStr = convStr +\
		    str[substrIndx:i] +\
		    frameToRTF[chr]
	else:
	  convStr = convStr +\
		    str[substrIndx:i] +\
		    '\\' +\
		    frameToRTF[chr]
        substrIndx = i + 1
    return(convStr + str[substrIndx:])
    
  # Given a string (potentially) with RTF special chars, produce Frame
  # output:

  def cnvRTFToFrame(self, str):

    self.errors = []
    self.warnings = []

    convStr = ''
    cntrStr = ''
    substrIndx = 0
    for i in range(len(str)):
      chr = str[i]
      # Beginning of a control sequence (e.g. "\{")?
      if chr == '\\':
	cntrStr = self.extractRTFControlStr(str, i)
        if cntrStr:	
          replacement = RTFToFrame[cntrStr]
	else:
	  replacement = ''
	# Copy the piece of string we traversed w/o finding control chars, and
	# add the translation of the cntr char:
	convStr = convStr +\
	   str[substrIndx:i] +\
           replacement
	# Advance the substr index to pt past the control seq:
	substrIndx = i + 1 + len(cntrStr)
      # End of control sequence processing
    return(convStr + str[substrIndx:])

  # Given a string and a start index, return a string with a RTF control
  # sequence, if it starts at the index. These are sequences of the form \{
  # or \\, or \lquote. Assumes that index points to the backslash that
  # starts the sequence:

  def extractRTFControlStr(self, str, i):
    if string.find(str, "{", i+1) == i+1:
      return("{")
    elif string.find(str, "}", i+1) == i+1:
      return("}")
    elif string.find(str, "lquote", i+1) == i+1:
      return("lquote ")
    elif string.find(str, "rquote", i+1) == i+1:
      return("rquote ")
    elif string.find(str, "rdblquote", i+1) == i+1:
      return("rdblquote ")
    elif string.find(str, "ldblquote", i+1) == i+1:
      return("ldblquote ")
    elif string.find(str, "endash", i+1) == i+1:
      return("endash ")
    elif string.find(str, "emdash", i+1) == i+1:
      return("emdash ")
    elif string.find(str, "\\", i+1) == i+1:
      return("\\")
    else:
      return('')

# Class and subclasses for converting attribute (field) names
# among Interbib and other systems (Refer, BibTeX, etc.):

class AttributeConverter:
  def __init__(self, pDict = {}):
    self.currentDict = pDict
  
  # Given a USMARC code, return an attribute that works on a given
  # information provider. Example: '100'->'au' for Dialog:

  def convertAttr(self, attr):

    if self.currentDict.has_key(attr):
      return(self.currentDict[attr])
    else:
      raise NonTranslatableAttribute, attr

  def reverseConvertAttr(self, name):

    for key in self.currentDict.keys():
      if self.currentDict[key] == name:
	return key
    raise NonTranslatableAttribute, name

import USMarc
class RawDLDocAttributeConverter(AttributeConverter):

  def __init__(self):
    self.currentDict = USMarc.USMarc
    self.reverseDict = USMarc.gEnglish2USMarc

  # Need to override the convert method, because USMarc.py's USMarc dictionnary
  # returns tuples, not single translations:

  def convertAttr(self,attr):
    if self.currentDict.has_key(attr):
      interbibAttr = string.lower(self.currentDict[attr][0])
      if interbibAttr == 'pagenum':
	interbibAttr = 'pages'
      return(interbibAttr)
    else:
      raise NonTranslatableAttribute, attr

  def reverseConvertAttr(self,attr):
    if self.reverseDict.has_key(attr):
      return(self.reverseDict[attr])
    else:
      raise NonTranslatableAttribute, attr

class ReferAttributeConverter(AttributeConverter):
  
  def __init__(self):

    # This is straight refer, except for %1: that's a custom
    # field that gets translated to be the field for link info.
    # Also: %0 is used to specify entry type if you wish:
    # book, article, etc. (see bibdb for proper classes):

    self.currentDict = {
      'A' : 'author', 'B': 'booktitle', 'C': 'address',
      'D': 'date', 'E': 'editor',
      'G': 'note', 'H': 'note', 'I': 'publisher',
      'J': 'journal', 'K': 'keywords', 'L': 'label',
      'N': 'number', 'O': 'note', 'F': keyFieldName,
      'P': 'pages', 'Q': 'institution', 'R': 'type',
      'S': 'series', 'T': 'title', 'U': urlFieldName,
      'V': 'volume', 'W': 'address', 'X': 'annote',
      '$': 'price',
      '*': 'copyright', '1': 'custom1', '2': 'custom2', 
      '3': 'custom3', '4': 'custom4'
      }

    # edition,

    self.reverseDict = {
      'annote': 'X', 'address': 'C', 
      'key': 'K', 'keywords': 'K', 
      'title': 'T', 'publisher': 'I', 'journal': 'J', 'date': 'D', 
      'price': '$', 'howpublished': 'W', 'number': 'N', 
      'booktitle': 'B', 'volume': 'V', urlFieldName : 'U', 
      'editor': 'E', 'copyright': '*', 'pages': 'P', 'series': 'S', 
      'label': 'L', 'organization': 'Q', keyFieldName: 'F',
      'institution': 'Q', 'header': 'H', 'entryType': '0', 
      'custom2': '2', 'custom3': '3', 'custom4': '4',
      'author': 'A', 'note': 'O', 'school': 'Q', 'type': 'R', 
      'year': 'date'
      }

  # Need to override the reverse convert method, because BibTeX <==> Refer is not
  # 1-to-1 in both directions. So we use a specially constructed dictionary
  # for each direction:

  def reverseConvertAttr(self,attr):

    if self.reverseDict.has_key(attr):
      return(self.reverseDict[attr])
    else:
      raise NonTranslatableAttribute, attr


class EndnoteAttributeConverter(ReferAttributeConverter):
  
  def __init__(self):

    # This is straight Endnote, except for %1: that's a custom
    # field that gets translated to be the field for link info.
    # Also: %0 is used to specify entry type if you wish:
    # book, article, etc. (see bibdb for proper classes):

    self.currentDict = {
      'X' : 'abstract', 'C' : 'address', '+' : 'affiliation', 
      'A' : 'author', 'B' : 'booktitle', 
      'S' : 'chapter', 'W' : 'copyright', 'L' : 'crossref', 
      '2' : 'custom2', '3' : 'custom3', '4' : 'custom4', 
      '8' : 'date', 'E' : 'editor', '0' : 'entryType', 
      'L' : 'governmentNumber', 'R' : 'howpublished', '?' : 'institution', 
      '@' : 'ISBN', 'J' : 'journal', 'M' : 'key', 
      'F' :  keyFieldName, 'K' : 'keywords', 'L' : 'label', 
      'G' : 'language', '@' : 'LCCN', '1' : linksFieldName, 
      'C' : 'location', '8' : 'month', 'Z' : 'note', 
      'N' : 'number', '?' : 'organization', 'P' : 'pages', 
      'I' : 'publisher',  
      'S' : 'series', 'T' : 'title', '2' : 'undefined', 
      'U' : urlFieldName, 'V' : 'volume', 'D' : 'year'
      }

    self.reverseDict = {
       'abstract' : 'X', 'address' : '+', 'affiliation' : '+', 
       'annote' : 'X', 'author' : 'A', 'booktitle' : 'B', 
       'chapter' : 'S', 'copyright' : 'W', 'crossref' : 'L', 
       'custom2' : '2', 'custom3' : '3', 'custom4' : '4', 
       'date' : '8', 'editor' : 'E', 'entryType' : '0', 
       'governmentNumber' : 'L', 'howpublished' : 'R', 
       'institution' : '?', 'ISBN' : '@', 'journal' : 'J', 
       'key' : 'M', keyFieldName : 'F', 'keywords' : 'K', 
       'label' : 'L', 'language' : 'G', 'LCCN' : '@', 
       linksFieldName : '1', 'location' : 'C', 'month' : '8', 
       'note' : 'Z', 'number' : 'N', 'organization' : '?', 
       'pages' : 'P', 'publisher' : 'I', 'references' : 'Z', 
       'school' : '?', 'series' : 'S', 'title' : 'T', 
       'undefined' : '2', urlFieldName : 'U', 'volume' : 'V', 'year' : 'D'
      }

class DialogAttributeConverter(AttributeConverter):

  def __init__(self):
    self.currentDict = {
      '020': 'bn', '022': 'sn', '100': 'au', '130': 'ti',
      '240': 'ti', '245': 'ti', '520': 'ab'
      }

class FolioAttributeConverter(AttributeConverter):

  def __init__(self):
    self.currentDict = {
      '100': 'PA', '130': 'AT', '240': 'AT', '245': 'AT', '520': 'AB',
      '504': 'DT'
      }

class WebAttributeConverter(AttributeConverter):

  def __init__(self):
    self.currentDict = {
      '130': 'title', '240': 'title', '245': 'title'
      }

# USMarc/Bib-1 attribute converter -- this is very rudimentary,
# and should be replaced by someone with a better idea of Bib-1
# and USMarc fields than I. - tom
class Bib1AttributeConverter(AttributeConverter):

  def __init__(self):
    self.currentDict = {
      '130': 4, '240': 4, '245': 4,      # Title
      '520': 62,                         # Abstract
      '100': 1003,                       # Author
      '945': 1010,                       # Content
      '650': 1035,                       # Subject
      '523': 31,                         # Publication Date
      '856': 2021                        # URL
      # Other Bib-1 we use (don't know USMarc) -- 2047 (Link), 5001 (Host),
      # 100001 (ADL Publisher), 100002 (ADL Scale)
    }
    self.reverseDict = {
      4: '245', 5: '245', 6: '245', 34: '245', 35: '245', 36: '245',
      37: '245', 38: '245', 39: '245', 40: '245', 41: '245', 42: '245', 
      43: '245', 44: '245',
      45: '520', 62: '520', 
      1003: '100', 1004: '100', 1005: '100', 1006: '100',
      1010: '945',
      1035: '650',
      30: '523', 31: '523',
      2021: '856'
      # Need 2047, 5001, 100001, 100002
    }

  def reverseConvertAttr(self,attr):
    if self.reverseDict.has_key(attr):
      return(self.reverseDict[attr])
    else:
      raise NonTranslatableAttribute, attr


# Class whose instances convert between Endnote and
# BibTeX record types (Conference Proceedings --> inproceedings, etc.):

class RecordTypeNameConverterEndnoteInterbib:
  
  def __init__(self):

    self.recordTypeDict = {
      'article' : 'Journal Article', 'book' : 'Book', 'booklet': 'Book',
      'conference' : 'Conference Proceedings', 'inbook' : 'Book Section',
      'incollection' : 'Book Section', 'inproceedings' : 'Conference Proceedings',
      'manual' : 'Book', 'mastersthesis' : 'Thesis', 'phdthesis' : 'Thesis',
      'misc' : 'Generic', 'proceedings' : 'Conference Proceedings', 'techreport' : 'Report',
      'unpublished' : 'Unpublished Work'
      }

    self.reverseRecordTypeDict = {
      'Journal Article' : 'article', 'Book' : 'book', 
      'Conference Proceedings' : 'inproceedings', 'Book Section' : 'incollection', 
      'Thesis' : 'misc', 'Generic' : 'misc', 'Conference Proceedings' : 'inproceedings', 
      'Report' : 'techreport', 'Unpublished Work' : 'unpublished'
      }

  def getInterBibRecTypeFromEndnote (self, endnoteRecType):
    try:
      return(self.reverseRecordTypeDict[endnoteRecType])
    except:
      raise NonTranslatableAttribute, endnoteRecType


  def getEndnoteRecTypeFromInterBib (self, interbibRecType):
    try:
      return(self.recordTypeDict[interbibRecType])
    except:
      raise NonTranslatableAttribute, interbibRecType

