################################################################################
#
# File:         bibdb.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibdb.py,v 1.6 2007/08/03 17:26:53 paepcke Exp $
# Description:  Database for bibliography entries
# Author:       Andreas Paepcke, Stanford University
# Created:      Mon Oct  9 08:08:01 1995
# Modified:     Thu Aug 02 12:53:14 2007 (Andreas Paepcke) paepcke@EMIL
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1995, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Sun Jul 22 07:48:35 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Fri Jul 20 11:36:15 2007 (Andreas Paepcke) paepcke@Shark.Stanford.EDU
#  Migrated from deprecated regex to re.
# Tue Apr  1 11:11:37 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added optional destination and anchor init arg in __init__ methods of WWWLink
#  and FrameLink
# Tue Apr  1 08:49:40 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  In Link class' __repr__ added closing '])' in printed repr.
# Mon Dec 30 09:55:27 1996 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added setting of char sets in importBibliography
# Fri Sep 20 11:36:04 1996 (Andreas Paepcke) paepcke@Walrus
#  Added setDbSharing method.
# Wed Sep 11 08:24:17 1996 (Andreas Paepcke) paepcke@Walrus
#  Changed default duplicate record behavior to autoresolve.
# Wed Aug 28 10:07:33 1996 (Andreas Paepcke) paepcke@Walrus
#  Added permission check in __delitem__
# Fri Jul 12 08:49:18 1996 (Andreas Paepcke) paepcke@Walrus
#  Added permissions
# Tue Jul  9 18:05:18 1996 (Andreas Paepcke) paepcke@Walrus
#  Made sharing with importers the default. Merging now done by importers
#  as they import.
# Tue Jul  9 09:06:39 1996 (Andreas Paepcke) paepcke@Walrus
#  Gave entries an additional interface to act as dicts.
# Mon Jul  8 19:42:34 1996 (Andreas Paepcke) paepcke@Walrus
#  Changed mergeRecords to (i) not return the length of shelves when
#  they are shared and the db was previously empty. That caused big delays
#  because the shelve needed to figure out its length. And (ii), for
#  empty self.entries when sharing is allowed, I now assign new entries
#  to self.entries even if dealing w/ a shelve. A comment said we shouldn't,
#  but I didn't see why not, and the merging was too slow.
# Sat Jul  6 15:29:33 1996 (Andreas Paepcke) paepcke@Stingray
#  Set shareEntries to 1 by default, set it to 0 after each bib import.
#  This was done to speed up bib import when only one import is done.
# Mon Jul  1 20:40:04 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  In describe: Made sure link objects were printed as strings.
# Thu Feb 22 07:39:43 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added __delattr__ to BibEntry class.
# Sun Jan  7 10:37:18 1996 (Andreas Paepcke) paepcke@AndreasHomePC
#  Added 'name' attr to DB class.
# Fri Dec 29 00:02:36 1995 (Andreas Paepcke) paepcke@AndreasHomePC
#  Modified mergeRecords to print warning if duplicate key recovery
#  and verbosity
# Thu Dec 28 23:39:03 1995 (Andreas Paepcke) paepcke@AndreasHomePC
#  Turned verbosity of BibEntry class off during import, because
#  importer already reports the conversion errors/warnings that
#  occur during entry field update.
# Mon Dec 18 10:38:15 1995 (Andreas Paepcke) paepcke@Stingray
#  Added verbosity to creation of importers.
# Thu Dec 14 10:43:55 1995 (Andreas Paepcke) paepcke@Stingray
#  Caused importBibliography to try and get bib file type from file extension
# Wed Dec  6 10:32:05 1995 (Andreas Paepcke) paepcke@Stingray
#  Added 'refer' as a legal bib system.
# Tue Dec  5 10:41:36 1995 (Andreas Paepcke) paepcke@Stingray
#  Replaced required_fields with requiredFields and optional_fields with
#  optionalFields.
# Tue Oct 31 09:17:54 1995 (Andreas Paepcke) paepcke@Stingray
#  Removed required_fields_one_of because it wasn't being used.
# Sun Oct 29 23:29:37 1995 (Andreas Paepcke) paepcke@Stingray
#  Added ability to pass dbStore optional arg to db creation. If present
#  this should be a shelve file name.
################################################################################

# Public Classes and Methods:
#
#   BibDB
#     attr: errors
#     attr: warnings
#     showErrors
#     showWarnings
#     importBibliography
#     reloadBibliography
#     setInputCharSet
#     setOutputCharSet
#     setNativeCharSet
#     sequence operators: +, db[k], del db[k], etc.
#     setPermissions

# To be done for more documentation:
#
#   - legal permissions: 'r', 'w', or 'rw'. Can create db w/ permissions or can
#     us setPermissions to change them afterwards. Even for read-only: can
#     import if not persistent. This allows opening of read-only InterBib dbs and
#     consolidating them into one in-memory.
#   - adding dbs through '+'
#   - importing new bibliographies through importBibliography
#   - persistence: pass an optional shelve file name to the constructor.
#     That will make the db persistent. Can import another persistent
#     db by importBibliography, with bibsystem='interbib'.
#   - len(db), db[<key>], del db[<key>], db[<key>] = <entry>
#   - db.items(), db.keys(), db.values(), db.has_key(<key>)
#   - showErrors, showWarnings
#   - verbosity
#   - setting char sets
#   - to control merge policies, you are allowed to set db.duplicateRecovery to:
#       - overwriteDuplicates
#       - autoResolveRecordConflicts
#       - duplicatesCauseErrors
#       - discardDuplicates
# - when importing existing bib entries into an empty db, dbs normally copy
#     all entries. This can be slow. You can set db.shareEntries to 1 to
#     cause sharing of those entries instead. But NOTE: If the imported
#     entries is a shelf (i.e. and InterBib db), any additional
#     bibliographies imported will be added to that shared bibliography db.
#     db.shareEntries is 1 by default. Every bibliography import will set it
#     to 0.
#   - links: - form with and without parens.
#            - Multiple target environments.
#            - comma between link infos is optional.
#            - Frame: gotolink <link-dest>  [<link-dest> is as speced in
#                                            special->marker: hypertext. Name
#            - special anchor field: 'citation'. If blank dest, the citation
#              in the doc will be linked to the reference in the bibl.
#              Else the citation will be linked to the dest.


#####################################################
#
# Global, Public Constants
#
#####################################################

# Behaviors when duplicate keys are detected during update:

duplicatesCauseErrors = 0       # This constant must be 0 for code to work!!
autoResolveRecordConflicts = 1  # If duplicate keys, invent a new key
overwriteDuplicates = 2         # If duplicate keys, overwrite
discardDuplicates = 3           # If duplicate keys, ignore the new ones.

# Import these after above constants 'cause they are used in bibimport:

import bibimport
import bibconvert
import bibutils
import bibsearch
import string
import re
import time
import types
import copy
import os
import shelve
import sys


#####################################################
#
# Exceptions
#
#####################################################

class DuplicateRecordKey(bibutils.InterbibException):
  pass

class PermissionDenied(bibutils.InterbibException):
  pass

#####################################################
#
# Random Constants, Semi-Private
#
#####################################################


#####################################################
#
# Class Definitions
#
#####################################################

# Permissions may be set to 'r' or 'w'

class BibDB:
  importer = 0
  supportedBibSystems = ['bibtex', 'DLRaw', 'interbib', 'refer', 'endnote']

  # Can pass in the name of a 'shelve' file:
  def __init__(self, dbStore = '', permissions='rw'):
    self.persistent = 0
    self.importedBefore = 0
    self.name = 'Unnamed'
    self.setPermissions(permissions)
    self.duplicateRecovery = autoResolveRecordConflicts
    self.defaultDuplicateRecovery = autoResolveRecordConflicts
    self.verbosity = 1
    self.sourceFile = ''
    self.sourceStr = ''
    self.sourceStr = None
    self.errors = []
    self.warnings = []
    self.updateIndexContinuously = 1
    self.searcher = bibsearch.bibSearch(dbClient=self)
    if dbStore:
      try:
	self.entries = {}
	self.fullFileName = os.path.expandvars(os.path.expanduser(dbStore))
	# For existing shelves, we can open w/ r or w or wr. But when we
	# need to create a shelve, that doesn't work:
	if os.path.isfile(self.fullFileName):
	  self.entries = shelve.open(self.fullFileName, self.shelvePermissions)
	else:
	  self.entries = shelve.open(self.fullFileName)
      except:
	try:
          self.errors.append("Could not open InterBib database '" + \
			     dbStore +\
			     "': " +\
			     sys.exc_value)
	except:
          self.errors.append("Could not open InterBib database '" + \
			     dbStore +\
			     "'")
        if self.verbosity:
	  bibutils.report('Error: ' + self.errors[-1])
	raise IOError, self.errors[-1]
      self.persistent = 1
      self.dbFile = dbStore
    else:
      self.entries = {}
      self.dbFile = ""

    # Allow sharing of bib entry dict structs w/ importer initially, since
    # most cases involved just a single import (apparently changed that
    # back to 0 to be save):
    self.shareEntries = 0

  # Normally, the first imported dataset will use the datastructure created
  # by the importer. Subsequent imports (into the same db) will therefore merge
  # into the db that was first imported. That's no big deal if the first db
  # is not persistent. But if you want the first set to stay virginal, you
  # can call this method and pass it a 0. That will force even the first
  # dataset to get its own data structure in memory. That's obviously expensive.

  def setDbSharing(self, allowSharing):
    self.importedBefore = not allowSharing
    self.shareEntries = allowSharing

  def setPermissions(self, permissions):
    # This first call will bomb for bad perm specs:
    self.shelvePermissions = self.getShelvePermissionsSpecs(permissions)
    self.permissions = permissions

  # Error-check a permissions str. return correct permissions to 
  # use for shelve opening:
  def getShelvePermissionsSpecs(self, permissions):
    if (permissions <> 'w') and (permissions <> 'r') and (permissions <> 'rw'):
      raise ValueError, "Permissions must be either 'r' or 'w' or 'rw'"
    if permissions == 'rw':
      return('w')
    else:
      return(permissions)

  # If inputCharSet is left unspecified, a good default will be picked.
  # Returns number of imported entries:

  def importBibliography(self,
			 sourceFile = "",
			 sourceStr = "",
			 sourceStruct = None,
			 verbosity = 1,
			 bibsystem = '',
			 duplicatePolicy = autoResolveRecordConflicts,
			 inputCharSet = None):
    if not bibsystem:
      if os.path.splitext(os.path.basename(sourceFile))[1] == '.ref':
        bibsystem = 'refer'
      elif os.path.splitext(os.path.basename(sourceFile))[1] == '.bib':
        bibsystem = 'bibtex'
      elif os.path.splitext(os.path.basename(sourceFile))[1] == '':
        bibsystem = 'interbib'
      else:
        bibsystem = 'bibtex'
    if bibsystem not in self.supportedBibSystems:
      self.errors.append('Unsupported bibliogaphy system: ' + bibsystem)
      if verbosity:
	bibutils.report('Error: ' + self.errors[-1])
      return(0)
    # If we are running verbose, the importer will repeat all the
    # bib entry's (i.e. the converter's) errors and warnings. So
    # we turn it off in the entry:
    BibEntry.verbosity = 0
    self.duplicateRecovery = duplicatePolicy
    self.sourceFile = os.path.expandvars(os.path.expanduser(sourceFile))
    self.sourceStr  = sourceStr
    self.sourceStruct = sourceStruct
    self.verbosity = verbosity
    self.errors = []
    self.warnings = []
    if inputCharSet:
      self.setInputCharSet(inputCharSet)
    if bibsystem == 'bibtex':
      # If caller didn't explicitly ask for a particular input char set,
      # set it to LaTeX:
      if not inputCharSet:
	self.setInputCharSet(bibconvert.LatexCharSet)
      self.importer = bibimport.BibtexImporter(fileName = self.sourceFile,
					       infoStr  = self.sourceStr,
					       infoStruct = self.sourceStruct,
					       verbosity = self.verbosity)
      numRecs = self.importer.importBibliography(verbose = verbosity,
					         duplicateRecovery = self.duplicateRecovery)
						 
    elif bibsystem == 'refer':
      # If caller didn't explicitly ask for a particular input char set,
      # set it to latin1:
      if not inputCharSet:
	self.setInputCharSet(bibconvert.Latin1CharSet)
      self.importer = bibimport.ReferImporter(fileName = self.sourceFile,
					      infoStr  = self.sourceStr,
					      infoStruct = self.sourceStruct,
					      verbosity = self.verbosity)			      
      numRecs = self.importer.importBibliography(self.sourceStruct,
					         duplicateRecovery = self.duplicateRecovery,
						 verbose = verbosity)

    elif bibsystem == 'endnote':
      # If caller didn't explicitly ask for a particular input char set,
      # set it to latin1:
      if not inputCharSet:
	self.setInputCharSet(bibconvert.Latin1CharSet)
      self.importer = bibimport.EndnoteImporter(fileName = self.sourceFile,
						infoStr  = self.sourceStr,
						infoStruct = self.sourceStruct,
						verbosity = self.verbosity)			      
      numRecs = self.importer.importBibliography(self.sourceStruct,
					         duplicateRecovery = self.duplicateRecovery,
						 verbose = verbosity)

    elif bibsystem == 'DLRaw':
      # If caller didn't explicitly ask for a particular input char set,
      # set it to latin1:
      if not inputCharSet:
	self.setInputCharSet(bibconvert.Latin1CharSet)
      self.importer = bibimport.HeuristicDLRawImporter(fileName = self.sourceFile,
						       infoStr  = self.sourceStr,
						       infoStruct = self.sourceStruct,
					               verbosity = self.verbosity)
      numRecs = self.importer.importBibliography(self.sourceStruct,
					         duplicateRecovery = self.duplicateRecovery,
						 verbose = verbosity)
    elif bibsystem == 'interbib':
      self.importer = bibimport.InterbibImporter(fileName = self.sourceFile,
						 permissions = self.shelvePermissions,
					         verbosity = self.verbosity)
      numRecs = self.importer.importBibliography(duplicateRecovery = self.duplicateRecovery,
						 verbose = verbosity)
      if not self.importedBefore:
        self.persistent = 1
	self.dbFile = self.sourceFile
    if self.importedBefore:
      # For the import/merging process: set the entries' verbosity to the
      # level we were told to do:
      BibEntry.verbosity = verbosity
      numRecs = self.mergeRecords(self.importer.entries)
    else:
      if self.persistent and not bibsystem == 'interbib':
	for key in self.importer.entries.keys():
          self[key] = self.importer.entries[key]
	# Flush the changes!!!
	self.entries.close()
	self.entries = shelve.open(self.fullFileName, self.shelvePermissions)
      else:
	self.entries = self.importer.entries
    self.importedBefore = 1
    self.errors = self.importer.errors
    self.warnings = self.importer.warnings
    return(numRecs)
      
  def reloadBibliography(self,
			 duplicatePolicy = autoResolveRecordConflicts):
    if not self.importer:
      self.errors.append('No previous import happened')
      if self.verbosity:
	bibutils.report('Error: ' + self.errors[-1])
      return(0)
    if duplicatePolicy == autoResolveRecordConflicts:
      self.importer.autoResolveRecordConflicts = 1
    else:
      self.importer.autoResolveRecordConflicts = 0
    self.importer.reloadBibliography(verbose = self.verbosity,
				     eraseAbbrevs = 1)
    self.errors = self.importer.errors
    self.warnings = self.importer.warnings
    oldduplicateRecovery = self.duplicateRecovery
    self.duplicateRecovery = overwriteDuplicates
    numRecs = self.mergeRecords(self.importer.entries)
    self.duplicateRecovery = oldduplicateRecovery
    return(numRecs)

  # NOTE: This will only return the number of records merged in if actual
  # merging was done.

  def mergeRecords(self, newEntries):
    if self.verbosity:
      startTime = time.time()
      bibutils.report('Merging entries into InterBib database...')
    # If we don't have any entries yet, we don't have to worry about duplicate
    # keys. The importers do that.
    if (len(self) == 0):
      # If we are allowed to share entries with the component we are
      # importing from, things are quick:
      if self.shareEntries:
        self.entries = newEntries
        # If the imported entry was a shelve, indicate that we are now
        # persistent:
        if type(self.entries) <> types.DictType:
  	  self.persistent = 1
      else:
	for key in newEntries.keys():
	  self.entries[key] = newEntries[key]
      if self.persistent:
	return('OK')
      else:
        return(len(self))
    mergedRecs = 0
    for key in newEntries.keys():
      if self.verbosity and (time.time() - startTime) > 5.0:
	bibutils.report('     Merged ' + str(mergedRecs) + ' records...')
	startTime = time.time()
      finalKey = key
      if self.entries.has_key(key):
	if self.duplicateRecovery == overwriteDuplicates:
	  # If we are maintaining an index dynamically, need to clean
	  # out the index entries of the entry about to be overwritten:
	  if self.updateIndexContinuously:
	    try:
	      self.searcher.removeEntryFromIndex(self.entries[key])
	    except:
	      self.openIndex()
	  self.warnings.append("Overwriting previous entry '" + key + "'")
    	  if self.verbosity:
            bibutils.report('Warning: ' + self.warnings[-1])
        else:
	  if self.duplicateRecovery == autoResolveRecordConflicts:
	    for i in range(100000):
	      newKey = key + '-' + str(i)
	      if not self.entries.has_key(newKey):
		self.warnings.append("Auto-resolved duplicate key '" +\
				     key +\
				     "' to new key '" +\
				     newKey +\
				     "'")
		if self.verbosity:
		  bibutils.report('Warning: ' + self.warnings[-1])
		finalKey = newKey
		break
	  elif self.duplicateRecovery == discardDuplicates:
	    self.warnings.append("Discarding duplicate entry '" + key + "'")
    	    if self.verbosity:
              bibutils.report('Warning: ' + self.warnings[-1])
	    continue
	  else:
	    self.errors.append("Key '" +\
				key +\
				"' is already in the bibliography database. Ignored.")
	    if self.verbosity:
	      raise DuplicateRecordKey, key
	    continue
      self[finalKey] = newEntries[key]
      self[finalKey].key = finalKey
      mergedRecs = mergedRecs + 1
    if self.persistent:
      # Flush the changes!!!
      self.entries.close()
      self.entries = shelve.open(self.fullFileName, self.shelvePermissions)
    return(mergedRecs)  
  
  def search(self, field='', value='', caseSensitive=0, exactMatch=0, useIndex=1):
    if not self.persistent:
      useIndex = 0
    if not self.searcher.isIndexOpen:
      self.openIndex()
    resDict = self.searcher.search(field = field,
			           value = value,
			           caseSensitive = caseSensitive,
			           exactMatch = exactMatch,
			           useIndex = useIndex)
    return(resDict)

  def openIndex(self):
    try:
      self.searcher.openIndexGivenDb()
    except bibsearch.IndexFileNotFound:
      self.createIndex()
      self.searcher.openIndexGivenDb()

  def createIndex(self):
    if not self.persistent:
      return
    self.searcher.createIndex(self.dbFile)
      
  def close(self):
    if self.persistent:
      self.entries.close()
      self.searcher.closeIndex()

  def merge(self, otherDb):
    return(self.mergeRecords(otherDb.entries))

  def __setattr__(self, name, value):
    if (name == 'inputCharSet')\
       or (name == 'outputCharSet')\
       or (name == 'nativeCharSet'):
      setattr(BibEntry, name, value)
    else:
      self.__dict__[name] = value

  def __add__(self, otherDb):
    newDb = copy.deepcopy(self)
    newDb.merge(otherDb)
    return(newDb)

  # Make it so that db['paep88'] will return a bib entry:
  def __getitem__(self, key):
    if 'r' in self.permissions:
      return(self.entries[key])
    else:
      raise PermissionDenied

  def __setitem__(self, key, value):
    if 'w' in self.permissions:
      self.entries[key] = value
      if self.persistent and self.updateIndexContinuously:
	try:
          self.searcher.addEntryToIndex(value)
	except bibsearch.IndexNotOpen:
	  self.openIndex()
    else:
      raise PermissionDenied

  def __delitem__(self, key):
    if 'w' in self.permissions:
      del self.entries[key]
    else:
      raise PermissionDenied

  def __len__(self):
    return(len(self.entries))
 
  def items(self):
    if 'r' in self.permissions:
      if self.persistent:
	res = []
	for key in self.entries.keys():
	  res.append((key, self.entries[key]))
	return(res)
      return(self.entries.items())
    else:
      raise PermissionDenied

  def keys(self):
    if 'r' in self.permissions:
      return(self.entries.keys())
    else:
      raise PermissionDenied
    
  def values(self):
    if 'r' in self.permissions:
      if self.persistent:
	res = []
	for key in self.entries.keys():
	  res.append(self.entries[key])
	return(res)
      return(self.entries.values())
    else:
      raise PermissionDenied
    
  def has_key(self, key):
    return(self.entries.has_key(key))

  def showErrors(self):
    if self.errors:
      for error in self.errors:
        bibutils.report(error)
    else:
      bibutils.report("No errors.")

  def showWarnings(self):
    if self.warnings:
      for warn in self.warnings:
        bibutils.report(warn)
    else:
      bibutils.report("No warnings.")

  def setInputCharSet(self, charSet):
    if charSet not in bibconvert.legalCharSets:
      raise bibconvert.UnsupportedCharSet, charSet
    BibEntry.inputCharSet = charSet
    
  def setOutputCharSet(self, charSet):
    if charSet not in bibconvert.legalCharSets:
      raise bibconvert.UnsupportedCharSet, charSet
    BibEntry.outputCharSet = charSet
    
  def setNativeCharSet(self, charSet):
    if charSet not in bibconvert.legalCharSets:
      raise bibconvert.UnsupportedCharSet, charSet
    BibEntry.nativeCharSet = charSet
    
# end class BibDB
      

# Class BibEntry is the parent class of all the specific bibliographic
# entries, like inproceedings, book, etc. Entries have three attribtes
# that can be set to control which character set conventions are to
# be used: inputCharSet specifies the character set used when setting
# bib fields. Attribute outputCharSet specifies the char set desired
# when bib fields are retrieved. Once these two attributes are set,
# input and output values are automatically converted accordingly.
# Attribute nativeCharSet controls which set is used internally. If
# you always put data in in the same format you want it to come out,
# you can gain efficiency by setting nativeCharSet to that char set.
# This avoids conversions.
# NOTE: These char set specs are class variables, not instance vars.
# NOTE: If your input format BibTex and your output format is also
# BibTex, you especially want to set inputCharSet to LatexCharSet,
# because you lose curlies that force capitalization.
# A list of legal char sets can be obtained
# through legalCharSets. Defaults are Latex for input and FrameMaker
# for output.
# The implementation uses __setattr__ and __getattr__ methods to handle
# the transparent translations.

class BibEntry:

  legalCharSets = bibconvert.legalCharSets
  inputCharSet  = bibconvert.LatexCharSet
  outputCharSet = bibconvert.FrameCharSet
  nativeCharSet = bibconvert.FrameCharSet
  warnings      = []
  errors        = []
  chrConverter  = bibconvert.StringPlatformConverter()
  verbosity     = 1

  def describe(self):
    keys = self.__dict__.keys()
    keys.sort()
    for key in keys:
      # Only print out the bibliographic fields, not the administrative ones:
      if key[0] == "_":
        # Strip the leading underscore from the field name:
	if key == '_' + bibconvert.linksFieldName:
	  bibutils.report(key[1:] + ":" + str(getattr(self,key)))
	else:
          bibutils.report(key[1:] + ": " + getattr(self,key))

  # The __setstate__ method is called when the entry is persistent and is
  # reconstituted upon extraction from the shelve. In Python1.4, the standard
  # unpickling process makes a new instance of BibEntry, and then calls
  # setattr on each key of the entry. Since these keys (in the shelve) have
  # their '_' pre-pended, our __setattr__ method will add *another* underscore,
  # which is wrong! So we do the re-setting of our dict here, without any
  # prepending of underscores, or conversion of values:

  def __setstate__(self, dict):
    for key in dict.keys():
      self.__dict__[key] = dict[key]

  # All field setting goes through here. The only built-in fields that are
  # supposed to be set are nativeCharSet, inputCharSet and outputCharSet.
  # We check the values for these for validity. All other fields are
  # assumed to be bibliographic fields. We put them into attributes with
  # the same name, but with '_' prepended, unless we are currently
  # unpickling a persistent entry (see comment with __setstate__). The
  # values are converted. When the attributes are retrieved under the name
  # they have been put in, they won't be found (because the underscore
  # won't be there. Therefore, __getattr__ will be called. It does the
  # retrieval and conversion.  We return a (possibly empty) array of
  # warnings which are strings:

  def __setattr__(self, name, value):
    if name == 'nativeCharSet' or\
       name == 'inputCharSet' or\
       name == 'outputCharSet':
      if value in self.legalCharSets:
	BibEntry.__dict__[name] = value
      else:
	raise UnsupportedCharSet, value
    elif name == 'warnings':
      # Warnings are another thing that may get set in the entry that isn't
      # a field:
      self.__dict__[name] = value
    elif (name == bibconvert.linksFieldName) or \
	 (name == bibconvert.urlFieldName) or \
	 (name == bibconvert.keyFieldName):
      # The links, url, and key fields are special. The first contains an array of link
      # objects, not strings. So there is no conversion. Same for the url field:
      # we don't want "-" in a URL to be converted to "--", for example:
      self.__dict__['_' + name] = value
    else:  # Assume we are setting a field of the entry
      self.warnings = []
      # Convert the string into native form and put it into a field _<name>:
      self.__dict__['_' + name] = self.cnvToNative(value)
      # The following return doesn't work: nothing gets returned. I guess
      # Python works that way...
      return(self.warnings)
      

  # Called when non-existent attr is being retrieved. This happens whenever
  # a normal attr, like 'author' is requested. Get the native string value
  # from the real field (e.g. _author) and return the result of converting
  # that to the output format we are set to:

  def __getattr__(self, name):
    if self.__dict__.has_key('_' + name):
      # A few fields are never character-converted. Screen those out:
      if (name == bibconvert.linksFieldName) or \
	 (name == bibconvert.urlFieldName) or \
	 (name == bibconvert.keyFieldName):
	return(getattr(self, '_' + name))
      else:
        return(self.cnvToOutput(getattr(self, '_' + name)))
    else:
      raise AttributeError, name

  def __delattr__(self, name):
    if self.__dict__.has_key('_' + name):
      del self.__dict__['_' + name]
    else:
      raise AttributeError, name

  # Give entries an additional dictionnary-like interface:

  def __getitem__(self, fldName):
    return(getattr(self, fldName))

  def __setitem__(self, fldName, value):
    setattr(self, fldName, value)

  def __delitem__(self, fldName):
    delattr(self, fldName)

  def __len__(self):
    length = 0
    for attr in self.__dict__.keys():
      if attr[0] == '_':
	length = length + 1
    return(length)

  def items(self):
    attrs = self.__dict__.keys()
    res = []
    for attr in attrs:
      if attr[0] == '_':
        res.append((attr[1:], getattr(self, attr)))
    return(res)

  def keys(self):
    theAttrs = self.__dict__.keys()
    res = []
    for attr in theAttrs:
      if attr[0] == '_':
	res.append(attr[1:])
    return(res)
    
  def values(self):
    attrs = self.__dict__.keys()
    res = []
    for attr in attrs:
      res.append(getattr(self, attr))
    return(res)
    
  def has_key(self, key):
    return(self.__dict__.has_key('_' + key))

  def cnvToNative(self, str):
    res = self.chrConverter.cnvSetToSet(str,
					BibEntry.inputCharSet,
					BibEntry.nativeCharSet)
    if self.chrConverter.errors:
      if BibEntry.verbosity:
	for error in self.chrConverter.errors:
	  bibutils.report('Error: ' + error)
      self.errors = self.errors + self.chrConverter.errors

    if self.chrConverter.warnings:
      if BibEntry.verbosity:
	for warn in self.chrConverter.warnings:
	  bibutils.report('Warning: ' + warn)
      self.warnings = self.warnings + self.chrConverter.warnings

    return(res)

  def cnvToOutput(self, str):

    res = self.chrConverter.cnvSetToSet(str,
					BibEntry.nativeCharSet,
					BibEntry.outputCharSet)
    if self.chrConverter.errors:
      if BibEntry.verbosity:
	for error in self.chrConverter.errors:
	  bibutils.report('Error: ' + error)
      self.errors = self.errors + self.chrConverter.errors

    if self.chrConverter.warnings:
      if BibEntry.verbosity:
	for warn in self.chrConverter.warnings:
	  bibutils.report('Warning: ' + warn)
      self.warnings = self.warnings + self.chrConverter.warnings

    return(res)

class article(BibEntry):
  recordType = 'article'
  requiredFields = ['author', 'title', 'journal', 'year']
  optionalFields = ['volume', 'number', 'pages', 'month', 'note']

class book(BibEntry):
  recordType = 'book'
  requiredFields = ['title', 'publisher', 'year']
  optionalFields = ['volume', 'series', 'address', 'edition', 'month', 'note']  

class booklet(BibEntry):
  recordType = 'booklet'
  requiredFields = ['title']
  optionalFields = ['author', 'howpublished', 'address', 'month', 'year',
		     'note'] 

class conference(BibEntry):
  recordType = 'conference'
  requiredFields = ['author', 'title', 'booktitle', 'year']
  optionalFields = ['editor', 'pages', 'organization', 'publisher', 'address',
		     'month', 'note'] 

class inbook(BibEntry):
  recordType = 'inbook'
  requiredFields = ['title', 'chapter', 'pages']
  optionalFields = ['volume', 'series', 'address', 'edition', 'month',
		     'note']  
  
class incollection(BibEntry):
  recordType = 'incollection'
  requiredFields = ['author', 'title', 'booktitle', 'publisher', 'year']
  optionalFields = ['editor', 'pages', 'organization', 'publisher', 'address',
		     'month', 'note'] 


class inproceedings(BibEntry):
  recordType = 'inproceedings'
  requiredFields = ['author', 'title', 'booktitle', 'year']
  optionalFields = ['editor', 'pages', 'organization', 'publisher', 'address',
		     'month', 'note'] 
  
class manual(BibEntry):
  recordType = 'manual'
  requiredFields = ['title']
  optionalFields = ['author', 'organization', 'address', 'edition', 'month',
		      'year', 'note'] 
  
class mastersthesis(BibEntry):
  recordType = 'mastersthesis'
  requiredFields = ['author', 'title', 'school', 'year']
  optionalFields = ['address', 'month', 'note'] 
  
class misc(BibEntry):
  recordType = 'misc'
  requiredFields = []
  optionalFields = ['author','title', 'howpublished', 'month', 'year',
		     'note'] 
  
class phdthesis(BibEntry):
  recordType = 'phdthesis'
  requiredFields = ['author', 'title', 'school', 'year']
  optionalFields = ['address', 'month', 'note'] 

class proceedings(BibEntry):
  recordType = 'proceedings'
  requiredFields = ['title', 'year']
  optionalFields = ['editor', 'publisher', 'organization', 'address',
		     'month', 'note'] 
  
class techreport(BibEntry):
  recordType = 'techreport'
  requiredFields = ['author', 'title', 'institution', 'year', 'number']
  optionalFields = ['type', 'address', 'month', 'note']   

class unpublished(BibEntry):
  recordType = 'unpublished'
  requiredFields = ['author', 'title', 'note']
  optionalFields = ['month', 'year'] 


def makeLinkObject(targetEnvironment):
  if targetEnvironment not in Link.legalTargetEnvironments:
    return(0)
  if targetEnvironment == 'www':
    return(WWWLink())
  elif targetEnvironment == 'frame':
    return(FrameLink())

# Class for holding links. Make a subclass for each new target environment.
# If you add __init__ or __setattr__ methods for those, you can do error checking
# on link destination info.
# The 'anchorField' in a link instance usually refers to a field within an entry.
# When the anchor field is set to 'citation', the link will be attached to the
# citation in a document, unless a field called 'citation' is also present in the
# entry.

class Link:
  legalTargetEnvironments = ['www', 'frame']
  anchorField = ''
  destination = ''

  def __repr__(self):
    res = 'Link('
    if self.__class__ == WWWLink:
      res = res + 'www'
    elif self.__class__ == FrameLink:
      res = res + 'frame'
    res = res + '|' + self.destination + '. Anchor:' + self.anchorField + ')'
    return(res)

  def ascii(self):
    res = '(' + self.anchorField + ':'
    if self.__class__ == WWWLink:
      res = res + 'www:'
    elif self.__class__ == FrameLink:
      res = res + 'frame:'
    res = res + self.destination + ')'
    return(res)

class WWWLink(Link):
  def __init__(self, dest='', anchor=''):
    self.system = 'www'
    self.destination = dest
    self.anchorField = anchor
    

class FrameLink(Link):
  def __init__(self, dest='', anchor=''):
    self.system = 'framemaker'
    self.destination = dest
    self.anchorField = anchor
