################################################################################
#
# File:         bibsearch.py
# RCS:          $Header: /u/testbed/CVSROOT/dldev/src/interbib2007/Scripts/bibsearch.py,v 1.1 2007/07/24 20:24:58 paepcke Exp $
# Description:  Search engine over InterBib databases
# Author:       Arturo Crespo, Stanford University
# Created:      Thu Jul  4 07:59:55 1996
# Modified:     Sun Jul 22 09:34:08 2007 (Andreas Paepcke) paepcke@SNUGGLES
# Language:     Python
# Package:      N/A
# Status:       Experimental (Do Not Distribute)
#
# (c) Copyright 1996, Stanford University, all rights reserved.
#
################################################################################
#
# Revisions:
#
# Sun Jul 22 07:58:26 2007 (Andreas Paepcke) paepcke@SNUGGLES
#  Converted to the new exception convention (no string exceptions; new calls for each)
# Fri Jul 20 11:38:08 2007 (Andreas Paepcke) paepcke@Shark.Stanford.EDU
#  Migrated from deprecated regex to re.
# Tue Mar 18 15:30:14 1997 (Andreas Paepcke) paepcke@Haddock.Stanford.EDU
#  Added removeEntryFromIndex
# Fri Jul 12 13:32:48 1996 (Andreas Paepcke) paepcke@Walrus
#  Changed chmod from 'a+w' to '644'
# Sun Jul  7 21:53:56 1996 (Andreas Paepcke) paepcke@Walrus
#  Completely redid the indexing structure and did index in
#  memory first.
# Sun Jul  7 14:59:39 1996 (Andreas Paepcke) paepcke@Stingray
#  Modified search to do OR of the words in a field.
# Sun Jul  7 14:59:02 1996 (Andreas Paepcke) paepcke@Stingray
#  Modified search to honor case sensitivity/exact matching even if
#  using the index
# Sat Jul  6 18:43:57 1996 (Andreas Paepcke) paepcke@Stingray
#  In createIndex: added percentage progress reporting for verbose mode.
# Sat Jul  6 18:43:30 1996 (Andreas Paepcke) paepcke@Stingray
#  Made output call bibutils.report to get flushing happening automatically
# Sat Jul  6 18:42:58 1996 (Andreas Paepcke) paepcke@Stingray
#  Modified so that createIndex could be used standalone, without any
#  calls to openIndex.
# Sat Jul  6 14:44:10 1996 (Andreas Paepcke) paepcke@Stingray
#  Added method closeIndex()
# Thu Jul  4 08:01:03 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  In createIndex: added code to strip any suffix from the file name
# Thu Jul  4 08:00:35 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Replaced import and use of POSIX by OS which is more portable.
# Thu Jul  4 08:00:23 1996 (Andreas Paepcke) paepcke@Phoebe.Stanford.EDU
#  Added standard header
################################################################################

#
# bibsearch.py
#
# Arturo Crespo  (crespo@cs.stanford.edu)
#
# Description:



import bibdb
import bibutils
import re
import os
import string
import shelve
import sys

#
# public methods:
#    __init__( verbosity )
#    __del__()
#    createIndex( dbFileName )
#    openIndex( dbFilename, dict, useIdx )
#    closeIndex()
#    search( fieldName, value, caseSensitive, exactMatch )
#
# Exceptions:
#    DBError
#    IndexNotOpen
#    IndexAccessError
#    IndexOld
#
# Description of public methods and usge:
#
# * __init__( verbosity )
#   constructor, verbosity level is by default 0 (i.e. very quiet)
# * __del__()
#   destructor
# * createIndex( dbFileName )
#   creates an index for the interbib file dbFileName, the index
#   is called dbFileName.idx
# * openIndex( dbFilename, dict )
#   which is in reality 2 overloaded methods:
#   * openIndex( dbFilename, {} )
#          opens an index for dbFilename.  Note that dbFilename doesn't
#          have the .idx extension
#   * openIndex( "", dictionary )
#          opens the dictionary passed as a parameter as an index.
# * closeIndex()
#   closes an index
#
# A very simple example of usage would be:
# 
#    bs=bibsearch.bibSearch(verbosity=ciVerbosity)
#    try:
#      bs.openIndex( dbname )
#    except IndexAccessError:
#      bs.createIndex( dbname )
#      bs.openIndex( dbname )
#   result = {}
#   result = bs.search( 'author', 'crespo' )
#

#####################################################
#
# Exceptions
#
#####################################################

class DBError(bibutils.InterbibException):
  pass

class IndexNotOpen(bibutils.InterbibException):
  pass

class IndexAccessError(bibutils.InterbibException):
  pass

class IndexFileNotFound(bibutils.InterbibException):
  pass

class IndexOld(bibutils.InterbibException):
  pass

#####################################################
#
# Class Definitions
#
#####################################################

class bibSearch:

  stopList = ["the","and","for"]

  def __init__(self, dbClient=None, verbosity=0) :
    self.verbosity = verbosity
    self.isIndexOpen =0;
    # Leave this '<> None' alone. Just 'if dbClient' has bootstrap problems w/ bibDb:
    if dbClient <> None:
      self.bibDb = dbClient
      self.bibliography = dbClient
    else:
      self.bibliography = None
      self.bibDb = None
    self.fieldsToIndex = ["title","author","key","date"]

  def __del__():
    if self.isIndexOpen:
      self.index.close()

  def  openIndex( self, dbFilename='', dict={}, useIdx=1, permissions='w'):
    if (dbFilename == ""):
      self.bibliography = dict
      self.index = {}
    else:
      self.getBibliography(dbFilename, permissions = permissions)
      self.index = {}

      if useIdx == 1:
	self.openIndexGivenDb(permissions=permissions)

    self.isIndexOpen  = 1

  # Used instead of openIndex if searcher is used from within a BibDB instance:
  def openIndexGivenDb(self, permissions='w'):
    if self.bibDb.dbFile:
      self.indexFileName = os.path.splitext(self.bibDb.dbFile)[0] + ".idx"
      if not os.path.isfile(self.indexFileName):
	raise IndexFileNotFound
      try:
        self.index = shelve.open(self.indexFileName, permissions)
      except os.error, info:
        raise IndexAccessError, info

    self.isIndexOpen  = 1

  def closeIndex(self):
    # If there is an open InterBib database, close it:
    if self.isIndexOpen:
      self.index.close()
    self.bibliography = None

  # This takes a bibliography database or a dict of key/bibEntry pairs such as is
  # returned by method search:
  def setBibliography(self, dictOfBibEntries):
    self.bibliography = dictOfBibEntries

  def getBibliography(self, dbFilename, permissions='w'):
    self.bibDb = bibdb.BibDB(dbStore=dbFilename,
			     permissions=permissions)
    self.bibliography = self.bibDb

  def search(self, field='', value='', caseSensitive=0, exactMatch=0, useIndex=1 ):

    if useIndex and not self.isIndexOpen:
      raise IndexNotOpen

    result = {}

    if ( field == "" ):
      return result

    if exactMatch and ( field == 'key'):
      for oneValue in string.split(value):
        if (oneValue in self.bibliography.keys()):
	  result[oneValue] = self.bibliography[oneValue]
      return result

    # Use the index if available:
    if field in self.fieldsToIndex and useIndex:
      if exactMatch:
	searchValues = [value]
      else:
	# Check all the words in value:
	tmp = string.split(value)
	searchValues = []
	# Remove duplicates:
	for word in tmp:
	  if word not in searchValues:
	    searchValues.append(word)

      for searchValue in searchValues:
	# If exact match, need to find first word that would actually be indexed:
	if exactMatch:
	  for exactWord in string.split(searchValues[0]):
	    if (exactWord in self.stopList) or (len(exactWord) < 3):
	      continue
	    else:
	      searchVal = string.lower(exactWord[0:4])
	else:
  	  searchVal = string.lower(searchValue[0:4])
        # If bibliography was set to a subset of the indexed collection,
	# some index hits may not really work, so put try around it:
	try:
	  for match in self.index[searchVal][field]:
	    entry = self.bibliography[match]
	    candidate = string.lower(bibutils.getEntryFldSafely(entry, field))
	    if (string.find(candidate,string.lower(searchValue))!=-1):
	      if caseSensitive and\
		 string.find(self.bibliography[match][field], searchValue) == -1:
		continue
	      result[match] = entry
	except KeyError:
	  continue
      return result
    else:
      # Non-indexed field or request not to use index:

      if exactMatch:
	repr = value
	if (caseSensitive):
	  reField = re.compile(repr)
	else:
	  reField = re.compile(repr,re.casefold)
	reFieldPatterns = [reField]

      else:
        reFieldPatterns = []
	for word in string.split(value):
	  repr = "\(.*\)" + word + "\(.*\)"
	  if (caseSensitive):
	    reField = re.compile(repr)
	  else:
	    reField = re.compile(repr,re.casefold)

	  reFieldPatterns.append(reField)

      for key in self.bibliography.keys():
	try:
	  candidate = self.bibliography[key][field]
	  for reField in reFieldPatterns:
	    if ( reField.match(candidate) == len(candidate) ):
	      result[key] = self.bibliography[key]
	except AttributeError:
	  pass

    return result


  def createIndex( self, dbFileName, adjustPermissions=1 ):

    # Remove any suffix from the file name:
    dbFileName = os.path.splitext(dbFileName)[0]
    try:
      if os.stat(dbFileName + '.idx'):
        os.system('rm ' + dbFileName + '.idx')
    except:
      pass
    if self.verbosity:
      bibutils.report("Creating Index...")
    self.index = shelve.open(dbFileName + ".idx")
    if not self.bibliography:
      self.getBibliography(dbFileName, permissions='r')
    bibKeys = self.bibliography.keys()
    if self.verbosity:
      keysToDo = len(bibKeys)
      feedbackGadget = bibutils.feedbackGadget(wholeJob = keysToDo,
					       reportPercentage = 1)
      bibutils.report('Beginning index phase...')
    memIndex = {}
    for key in bibKeys:
      if self.verbosity:
        feedbackGadget.tick(1)
      entry = self.bibliography[key]
      for fieldName in self.fieldsToIndex:
	fieldValue = string.lower(bibutils.getEntryFldSafely(entry, fieldName))
	if fieldName == 'author':
	  tokens = string.split(fieldValue," \||")
	elif fieldName == 'key':
	  tokens = [fieldValue]
	else:
	  tokens = string.split(fieldValue, "[^A-Za-z0-9]*")
  	for token in tokens:
  	  if not(token in self.stopList) and (len(token)>=3):
	    word = token[0:4]
	    try:
	      memIndex[word][fieldName].append(key)
	    except:
	      memIndex[word] = {}
	      for aField in self.fieldsToIndex:
	        memIndex[word][aField] = []
	      memIndex[word][fieldName].append(key)

    for word in memIndex.keys():
      self.index[word] = memIndex[word]
    # Make sure everything is flushed:
    self.index.close()
    if adjustPermissions:
      try:
        os.system('chmod 644 ' + dbFileName + '.idx')
      except:
	pass
    self.indexFileName = dbFileName + ".idx"
    self.index = shelve.open(self.indexFileName)

  # Given a bibEntry, index it incrementally into the existing index.
  # Note: shelves are only written to at the top level. We cannot add
  # information into the structure *within* one value in a shelve. Instead,
  # we need to read the value out of the shelve, modify it and write the
  # whole structure back. Therefore the 'wordDicts' structure below.

  def addEntryToIndex(self, entry):
    if not self.isIndexOpen:
      raise IndexNotOpen
    wordDicts = {}
    for fieldName in self.fieldsToIndex:
      fieldValue = string.lower(bibutils.getEntryFldSafely(entry, fieldName))
      if fieldName == 'author':
        tokens = string.split(fieldValue," \||")
      elif fieldName == 'key':
	tokens = [fieldValue]
      else:
        tokens = string.split(fieldValue, "[^A-Za-z0-9]*")
      for token in tokens:
        if not(token in self.stopList) and (len(token)>=3):
          word = token[0:4]
	  if wordDicts.has_key(word) and entry.key not in wordDicts[word][fieldName]:
	    wordDicts[word][fieldName].append(entry.key)
	    continue
          try:
	    wordDict = self.index[word]
	    if entry.key not in wordDict[fieldName]:
	      wordDict[fieldName].append(entry.key)
            wordDicts[word] = wordDict
          except:
            wordDict = {}
            for aField in self.fieldsToIndex:
              wordDict[aField] = []
            wordDict[fieldName].append(entry.key)
	    wordDicts[word] = wordDict
    for word in wordDicts.keys():
      self.index[word] = wordDicts[word]
    # Flush the changes:      
    self.index.close()
    self.index = shelve.open(self.indexFileName)

  # Given a bibEntry, remove references to it from the index.
  # Note: shelves are only written to at the top level. We cannot add
  # information into the structure *within* one value in a shelve. Instead,
  # we need to read the value out of the shelve, modify it and write the
  # whole structure back. Therefore the 'wordDicts' structure below.
  # Note: this code is very similar to addEntryToIndex above. They may be
  # combined. But on the other hand, we want few if/else's in the code for
  # speed. Mostly, addEntryToIndex works and I don't want to fuzz w/ it...

  def removeEntryFromIndex(self, entry):
    if not self.isIndexOpen:
      raise IndexNotOpen
    wordDicts = {}
    for fieldName in self.fieldsToIndex:
      fieldValue = string.lower(bibutils.getEntryFldSafely(entry, fieldName))
      if fieldName == 'author':
        tokens = string.split(fieldValue," \||")
      elif fieldName == 'key':
	tokens = [fieldValue]
      else:
        tokens = string.split(fieldValue, "[^A-Za-z0-9]*")
      for token in tokens:
        if not(token in self.stopList) and (len(token)>=3):
          word = token[0:4]
	  if wordDicts.has_key(word) and entry.key in wordDicts[word][fieldName]:
	    wordDicts[word][fieldName].remove(entry.key)
	    continue
          try:
	    wordDict = self.index[word]
	    if entry.key in wordDict[fieldName]:
	      wordDict[fieldName].remove(entry.key)
            wordDicts[word] = wordDict
          except:
	    # Shouldn't happen if the entry is really in the index, but...
	    continue

    for word in wordDicts.keys():
      # If a word was only in the entry we deleted, remove the word from the index:
      wordEmpty = 1
      for fld in self.fieldsToIndex:
	if wordDicts[word][fld]:
	  wordEmpty = 0
	  break
      if wordEmpty:
	del self.index[word]
      else:
        self.index[word] = wordDicts[word]
    # Flush the changes:
    self.index.close()
    self.index = shelve.open(self.indexFileName)

  # Assuming the index is open, print its contents:

  def displayIndex(self):
    try:
      keys = self.index.keys()
    except:
      raise IndexAccessError, "Maybe index not open?"
    keys.sort()
    for key in keys:
      print key + ": " + str(self.index[key])
